# Copyright (c) Acconeer AB, 2023-2025
# All rights reserved

from __future__ import annotations

import enum
import warnings
from typing import Any, Dict, Optional, Tuple

import attrs
import h5py
from attributes_doc import attributes_doc

from acconeer.exptool import a121, opser
from acconeer.exptool import type_migration as tm
from acconeer.exptool.a121._core import utils
from acconeer.exptool.a121._h5_utils import _create_h5_string_dataset
from acconeer.exptool.a121.algo import (
    AlgoConfigBase,
    Controller,
    PeakSortingMethod,
    ReflectorShape,
)
from acconeer.exptool.a121.algo._utils import ENVELOPE_FWHM_M
from acconeer.exptool.a121.algo.distance import (
    PRF_REMOVED_ET_VERSION,
    Detector,
    DetectorConfig,
    DetectorContext,
    DetectorResult,
    ThresholdMethod,
    _DetectorConfig_v0,
)
from acconeer.exptool.a121.algo.distance._detector import detector_context_timeline

from ._processor import (
    Processor,
    ProcessorConfig,
    ProcessorExtraResult,
    ProcessorLevelStatus,
    ProcessorResult,
)


PARTIAL_RANGE_FWHM_FACTOR = 2.0  # Sets the minimal partial range in number of FWHM widths


class RangeMode(enum.Enum):
    """Tank Level range mode."""

    FULL_RANGE = enum.auto()
    PARTIAL_RANGE = enum.auto()


@attributes_doc
@attrs.mutable(kw_only=True)
class RefAppConfig(AlgoConfigBase):
    start_m: float = attrs.field(default=0.03)
    """Start of measurement range."""

    end_m: float = attrs.field(default=0.5)
    """End of measurement range."""

    max_step_length: Optional[int] = attrs.field(default=None)
    """If set, limits the step length.

    If no argument is provided, the step length is automatically calculated based on the profile.

    Reducing the step length increases SNR through more efficient distance filtering, while
    increasing the measurement time and the processing load.
    """

    max_profile: a121.Profile = attrs.field(default=a121.Profile.PROFILE_5, converter=a121.Profile)
    """Specifies the longest allowed profile.

    If no argument is provided, the highest possible profile without interference of direct
    leakage is used to maximize SNR.

    A lower profile improves the radial resolution.
    """

    close_range_leakage_cancellation: bool = attrs.field(default=False)
    """Enable close range leakage cancellation logic.

    Close range leakage cancellation refers to the process of measuring close to the
    sensor(<100mm) by first characterizing the direct leakage, and then subtracting it
    from the measured sweep in order to isolate the signal component of interest.

    The close range leakage cancellation process requires the sensor to be installed in its
    intended geometry with free space in front of the sensor during detector calibration.
    """

    level_tracking_active: bool = attrs.field(default=False)
    """Track level and measure only a smaller partial range to save power"""

    partial_tracking_range_m: float = attrs.field(default=1.0)
    """Minimum partial range window length used during level tracking"""

    signal_quality: float = attrs.field(default=15.0)
    """Signal quality (dB).

    High quality equals higher HWAAS and better SNR but increases power consumption."""

    threshold_method: ThresholdMethod = attrs.field(
        default=ThresholdMethod.CFAR,
        converter=ThresholdMethod,
    )
    """Threshold method"""

    peaksorting_method: PeakSortingMethod = attrs.field(
        default=PeakSortingMethod.STRONGEST,
        converter=PeakSortingMethod,
    )
    """Sorting method of estimated distances.

    The distance estimates are sorted according to the selected strategy, before being return
    by th application.
    """

    reflector_shape: ReflectorShape = attrs.field(
        default=ReflectorShape.GENERIC,
        converter=ReflectorShape,
    )
    """Reflector shape."""

    num_frames_in_recorded_threshold: int = attrs.field(default=100)
    """Number of frames used when calibrating threshold.

    A lower number reduce calibration time and a higher number results in a more statistically
    significant threshold.
    """

    fixed_threshold_value: float = attrs.field(default=100.0)
    """Value of fixed amplitude threshold."""

    fixed_strength_threshold_value: float = attrs.field(default=0.0)
    """Value of fixed strength threshold."""

    threshold_sensitivity: float = attrs.field(default=0.5)
    """Sensitivity of threshold.

    High sensitivity equals low detection threshold, low sensitivity equals high detection
    threshold."""

    update_rate: Optional[float] = attrs.field(default=50.0)
    """Sets the detector update rate."""

    median_filter_length: int = attrs.field(default=5)
    """Length of the median filter used to improve robustness of the result."""

    num_medians_to_average: int = attrs.field(default=1)
    """Number of medians averaged to obtain the final level."""

    @start_m.validator
    def _(self, _: Any, value: float) -> None:
        if value < Detector.MIN_DIST_M:
            msg = f"Cannot start measurements closer than {Detector.MIN_DIST_M}m"
            raise ValueError(msg)

    @end_m.validator
    def _(self, _: Any, value: float) -> None:
        if value > Detector.MAX_DIST_M:
            msg = f"Cannot measure further than {Detector.MAX_DIST_M}m"
            raise ValueError(msg)

    def _collect_validation_results(self) -> list[a121.ValidationResult]:
        validation_results: list[a121.ValidationResult] = []

        if self.threshold_method == ThresholdMethod.RECORDED and self.level_tracking_active:
            validation_results.append(
                a121.ValidationError(
                    self,
                    "threshold_method",
                    "Cannot use recorded threshold when level tracking is activated.",
                )
            )
        if self.close_range_leakage_cancellation and self.level_tracking_active:
            validation_results.append(
                a121.ValidationError(
                    self,
                    "close_range_leakage_cancellation",
                    "Cannot use close range leakage cancellation when level tracking is activated.",
                )
            )
        if (
            self.level_tracking_active
            and self.partial_tracking_range_m is not None
            and self.partial_tracking_range_m
            <= ENVELOPE_FWHM_M[self.max_profile] * PARTIAL_RANGE_FWHM_FACTOR
        ):
            validation_results.append(
                a121.ValidationError(
                    self,
                    "partial_tracking_range_m",
                    (
                        "Partial tracking range must be larger than "
                        f"{ENVELOPE_FWHM_M[self.max_profile] * PARTIAL_RANGE_FWHM_FACTOR:.3f} "
                        "for the selected max_profile."
                    ),
                )
            )
        if (
            self.level_tracking_active
            and self.partial_tracking_range_m is not None
            and self.partial_tracking_range_m >= (self.end_m - self.start_m)
        ):
            validation_results.append(
                a121.ValidationError(
                    self,
                    "partial_tracking_range_m",
                    "Partial tracking range must be smaller than (end_m - start_m).",
                )
            )
        if self.end_m < self.start_m:
            validation_results.append(
                a121.ValidationError(
                    self,
                    "start_m",
                    "Must be smaller than 'Range end'",
                )
            )

            validation_results.append(
                a121.ValidationError(
                    self,
                    "end_m",
                    "Must be greater than 'Range start'",
                )
            )

        if self.max_step_length is not None and (
            not utils.is_divisor_of(24, self.max_step_length)
            and not utils.is_multiple_of(24, self.max_step_length)
        ):
            valid_step_length = next(
                sl
                for sl in range(self.max_step_length, 0, -1)
                if utils.is_divisor_of(24, sl) or utils.is_multiple_of(24, sl)
            )
            validation_results.append(
                a121.ValidationWarning(
                    self,
                    "max_step_length",
                    "Actual step length will be rounded down "
                    + f"to the closest valid step length ({valid_step_length}).",
                )
            )

        return validation_results

    def to_detector_config(self) -> DetectorConfig:
        """Nominal full range detector config"""

        return DetectorConfig(
            start_m=self.start_m - 0.015,
            end_m=min(self.end_m * 1.05, 23.0),
            max_step_length=self.max_step_length,
            max_profile=self.max_profile,
            close_range_leakage_cancellation=self.close_range_leakage_cancellation,
            signal_quality=self.signal_quality,
            threshold_method=self.threshold_method,
            peaksorting_method=self.peaksorting_method,
            reflector_shape=self.reflector_shape,
            num_frames_in_recorded_threshold=self.num_frames_in_recorded_threshold,
            fixed_strength_threshold_value=self.fixed_strength_threshold_value,
            fixed_threshold_value=self.fixed_threshold_value,
            threshold_sensitivity=self.threshold_sensitivity,
            update_rate=None,
        )

    def to_detector_level_tracking_config(self, level: float) -> DetectorConfig:
        """Level tracking detector config."""
        full_range_config = self.to_detector_config()
        distance = self.end_m - level
        if self.partial_tracking_range_m is not None:
            partial_range_width = max(
                self.partial_tracking_range_m,
                ENVELOPE_FWHM_M[self.max_profile] * PARTIAL_RANGE_FWHM_FACTOR,
            )
        else:
            partial_range_width = ENVELOPE_FWHM_M[self.max_profile] * PARTIAL_RANGE_FWHM_FACTOR
        sub_start = max(distance - partial_range_width / 2, full_range_config.start_m)
        sub_end = min(distance + partial_range_width / 2, full_range_config.end_m)

        return DetectorConfig(
            start_m=sub_start,
            end_m=sub_end,
            max_step_length=self.max_step_length,
            max_profile=self.max_profile,
            close_range_leakage_cancellation=self.close_range_leakage_cancellation,
            signal_quality=self.signal_quality,
            threshold_method=self.threshold_method,
            peaksorting_method=self.peaksorting_method,
            reflector_shape=self.reflector_shape,
            num_frames_in_recorded_threshold=self.num_frames_in_recorded_threshold,
            fixed_strength_threshold_value=self.fixed_strength_threshold_value,
            fixed_threshold_value=self.fixed_threshold_value,
            threshold_sensitivity=self.threshold_sensitivity,
            update_rate=None,
        )


@attrs.frozen(kw_only=True)
class RefAppExtraResult:
    processor_extra_result: ProcessorExtraResult
    detector_result: Dict[int, DetectorResult]


RefAppContext = DetectorContext
ref_app_context_timeline = detector_context_timeline


@attrs.frozen(kw_only=True)
class RefAppResult:
    peak_detected: Optional[bool]
    """True if a peak (level) is detected, False if no peak is
    detected, or None if a result is not available."""
    peak_status: Optional[ProcessorLevelStatus]
    """Status assigned to the detected peak."""
    level: Optional[float]
    """Liquid level relative to the base of the tank."""
    extra_result: RefAppExtraResult
    """Extra result: Only used for the plots in the GUI."""


class RefApp(Controller[RefAppConfig, RefAppResult]):
    def __init__(
        self,
        *,
        client: a121.Client,
        sensor_id: int,
        config: RefAppConfig,
        context: Optional[RefAppContext] = None,
    ) -> None:
        super().__init__(client=client, config=config)
        self.sensor_id = sensor_id
        self.range_mode: RangeMode = RangeMode.FULL_RANGE
        self.detector_config = self.config.to_detector_config()

        self._detector = Detector(
            client=self.client,
            sensor_ids=[self.sensor_id],
            detector_config=self.detector_config,
            context=context,
        )

        processor_config = ProcessorConfig(
            median_filter_length=self.config.median_filter_length,
            num_medians_to_average=self.config.num_medians_to_average,
            tank_range_start_m=self.config.start_m,
            tank_range_end_m=self.config.end_m,
        )

        self._processor = Processor(processor_config)

        self.started = False
        self.context = context

    def calibrate(self) -> None:
        self._detector.calibrate_detector()

    def start(
        self, recorder: Optional[a121.Recorder] = None, algo_group: Optional[h5py.Group] = None
    ) -> None:
        if self.started:
            msg = "Already started"
            raise RuntimeError(msg)

        if recorder is not None:
            if isinstance(recorder, a121.H5Recorder):
                algo_group = recorder.require_algo_group("tank_level")
                _record_algo_data(algo_group, self.sensor_id, self.config, self._detector.context)
            else:
                # Should never happen as we currently only have the H5Recorder
                warnings.warn("Will not save algo data")

        # Detector start() called with recorder and algo group input.
        # Since the detector is swapped and/or stopped during runtime without
        # recorder and algo group input a side effect is that, only the
        # algo group and context for the initial full range config is saved to the record.
        self._detector.start(recorder=recorder, _algo_group=algo_group)

        self.started = True

    def _get_next_mean_median(self) -> Tuple[ProcessorResult, RefAppExtraResult]:
        current_peak_status = None
        # fetch new data until processor has accumulated enough data to give result.
        # get_next() will be called {median_filter_length * num_medians_to_average} times.
        while current_peak_status is None:
            result = self._detector.get_next()
            processor_result = self._processor.process(
                detector_result=result,
                detector_start_m=self._detector.config.start_m,
            )
            current_peak_status = processor_result.peak_status

        ref_app_extra_result = RefAppExtraResult(
            processor_extra_result=processor_result.extra_result, detector_result=result
        )
        return (processor_result, ref_app_extra_result)

    def get_next(self) -> RefAppResult:
        if not self.started:
            msg = "Not started"
            raise RuntimeError(msg)

        if not self._detector.started:
            self._detector.calibrate_detector()
            self._detector.start(recorder=None, _algo_group=None)
        (processor_result, ref_app_extra_result) = self._get_next_mean_median()

        if self.config.update_rate is not None:
            self._detector.stop_detector()

        # Level tracking (partial/full range) state machine.
        if self.range_mode == RangeMode.FULL_RANGE:
            # Do nothing while in full range mode. I.e sensor range is kept constant in this
            # mode as full range
            if (
                processor_result.peak_status == ProcessorLevelStatus.IN_RANGE
                and processor_result.filtered_level is not None
                and self.config.level_tracking_active
            ):
                # change to tracking mode. I.e partial range
                detector_config = self.config.to_detector_level_tracking_config(
                    processor_result.filtered_level
                )
                self._swap_detector(detector_config)
                self.range_mode = RangeMode.PARTIAL_RANGE
        elif self.range_mode == RangeMode.PARTIAL_RANGE:
            # update config for level tracking
            if (
                processor_result.peak_status != ProcessorLevelStatus.IN_RANGE
                or processor_result.filtered_level is None
            ):
                # No peak detected (NO_DETECTION) or outside range (OUT_OF_RANGE or OVERFLOW).
                # Go to full range mode
                detector_config = self.config.to_detector_config()
                self._swap_detector(detector_config)
                self.range_mode = RangeMode.FULL_RANGE
            else:
                detector_config = self.config.to_detector_level_tracking_config(
                    processor_result.filtered_level
                )
                self._swap_detector(detector_config)

        return RefAppResult(
            peak_detected=processor_result.peak_detected,
            peak_status=processor_result.peak_status,
            level=processor_result.filtered_level,
            extra_result=ref_app_extra_result,
        )

    def update_config(self, config: RefAppConfig) -> None:
        raise NotImplementedError

    def stop(self) -> Any:
        if not self.started:
            msg = "Already stopped"
            raise RuntimeError(msg)

        if self._detector.started:
            recorder_result = self._detector.stop()
        else:
            recorder_result = self._detector.stop_recorder()

        self.started = False

        return recorder_result

    def _swap_detector(self, config: DetectorConfig) -> None:
        """Create new detector object. No recorder is sent to detector.
        The recorder is however attached to the client at initialization of the ref app."""
        if self._detector.started:
            self._detector.stop_detector()
        self._detector = Detector(
            client=self.client,
            sensor_ids=[self.sensor_id],
            detector_config=config,
            context=self.context,
        )


def _record_algo_data(
    algo_group: h5py.Group, sensor_id: int, config: RefAppConfig, context: RefAppContext
) -> None:
    algo_group.create_dataset("sensor_id", data=sensor_id, track_times=False)

    _create_h5_string_dataset(algo_group, "config", config.to_json())

    context_group = algo_group.create_group("tank_level_context")
    opser.serialize(context, context_group)


def _load_algo_data(algo_group: h5py.Group) -> Tuple[int, RefAppConfig, RefAppContext]:
    sensor_id = int(algo_group["sensor_id"][()])

    try:
        config = ref_app_config_timeline.migrate(algo_group["config"][()].decode())
    except tm.core.MigrationErrorGroup as exc:
        msg = ""
        match = exc.subgroup(BadMigrationPathError)
        if match is not None:
            # Add more details from exception thrown
            for exc_arg in match.exceptions[0].args:
                if isinstance(exc_arg, str):
                    print(exc_arg)
                    msg += f", {exc_arg}"

        raise TypeError(msg) from exc

    context_group = algo_group["tank_level_context"]
    tank_level_context = ref_app_context_timeline.migrate(context_group)

    return sensor_id, config, tank_level_context


@attrs.mutable
@attrs.mutable(kw_only=True)
class _RefAppConfig_v1(AlgoConfigBase):
    start_m: float = attrs.field(default=0.03)
    end_m: float = attrs.field(default=0.5)
    max_step_length: Optional[int] = attrs.field(default=None)
    max_profile: a121.Profile = attrs.field(default=a121.Profile.PROFILE_5, converter=a121.Profile)
    close_range_leakage_cancellation: bool = attrs.field(default=False)
    signal_quality: float = attrs.field(default=15.0)
    threshold_method: ThresholdMethod = attrs.field(
        default=ThresholdMethod.CFAR,
        converter=ThresholdMethod,
    )
    peaksorting_method: PeakSortingMethod = attrs.field(
        default=PeakSortingMethod.STRONGEST,
        converter=PeakSortingMethod,
    )
    reflector_shape: ReflectorShape = attrs.field(
        default=ReflectorShape.GENERIC,
        converter=ReflectorShape,
    )
    num_frames_in_recorded_threshold: int = attrs.field(default=100)
    fixed_threshold_value: float = attrs.field(default=100.0)
    fixed_strength_threshold_value: float = attrs.field(default=0.0)
    threshold_sensitivity: float = attrs.field(default=0.5)
    update_rate: Optional[float] = attrs.field(default=50.0)
    median_filter_length: int = attrs.field(default=5)
    num_medians_to_average: int = attrs.field(default=1)

    def _collect_validation_results(self) -> list[a121.ValidationResult]:
        return []

    def migrate(self) -> RefAppConfig:
        return RefAppConfig(
            start_m=self.start_m,
            end_m=self.end_m,
            max_step_length=self.max_step_length,
            max_profile=self.max_profile,
            close_range_leakage_cancellation=self.close_range_leakage_cancellation,
            signal_quality=self.signal_quality,
            update_rate=self.update_rate,
            median_filter_length=self.median_filter_length,
            num_medians_to_average=self.num_medians_to_average,
            threshold_method=self.threshold_method,
            reflector_shape=self.reflector_shape,
            peaksorting_method=self.peaksorting_method,
            num_frames_in_recorded_threshold=self.num_frames_in_recorded_threshold,
            fixed_threshold_value=self.fixed_threshold_value,  # float
            fixed_strength_threshold_value=self.fixed_strength_threshold_value,  # float
            threshold_sensitivity=self.threshold_sensitivity,  # float
            level_tracking_active=False,
            partial_tracking_range_m=0.0,  # not used
        )


@attrs.mutable
@attrs.mutable(kw_only=True)
class _RefAppConfig_v0(_DetectorConfig_v0):
    start_m: float = attrs.field(default=0.03)
    end_m: float = attrs.field(default=0.5)
    median_filter_length: int = attrs.field(default=5)
    num_medians_to_average: int = attrs.field(default=1)
    close_range_leakage_cancellation: bool = attrs.field(default=False)

    def to_detector_config(self) -> RefAppConfig:
        return RefAppConfig(
            start_m=self.start_m - 0.015,
            end_m=min(self.end_m * 1.05, 23.0),
            max_step_length=self.max_step_length,
            max_profile=self.max_profile,
            close_range_leakage_cancellation=self.close_range_leakage_cancellation,
            signal_quality=self.signal_quality,
            threshold_method=self.threshold_method,
            peaksorting_method=self.peaksorting_method,
            reflector_shape=self.reflector_shape,
            num_frames_in_recorded_threshold=self.num_frames_in_recorded_threshold,
            fixed_strength_threshold_value=self.fixed_strength_threshold_value,
            fixed_threshold_value=self.fixed_threshold_value,
            threshold_sensitivity=self.threshold_sensitivity,
            update_rate=self.update_rate,
        )

    def _collect_validation_results(self) -> list[a121.ValidationResult]:
        return []


class BadMigrationPathError(Exception): ...


def always_raise_an_error_when_migrating_v0(_: _DetectorConfig_v0) -> _RefAppConfig_v1:
    msg = f"Try opening the file in an earlier version of ET <= {PRF_REMOVED_ET_VERSION}"
    raise BadMigrationPathError(msg)


ref_app_config_timeline = (
    tm.start(_RefAppConfig_v0)
    .load(str, _RefAppConfig_v0.from_json, fail=[TypeError])
    .nop()
    .epoch(_RefAppConfig_v1, always_raise_an_error_when_migrating_v0, fail=[BadMigrationPathError])
    .load(str, _RefAppConfig_v1.from_json, fail=[TypeError])
    .nop()
    .epoch(RefAppConfig, _RefAppConfig_v1.migrate, fail=[])
    .load(str, RefAppConfig.from_json, fail=[TypeError])
)



# Copyright (c) Acconeer AB, 2022-2025
# All rights reserved

from __future__ import annotations

import warnings
from typing import Any, Optional, Tuple

import attrs
import h5py
import numpy as np
import numpy.typing as npt
from attributes_doc import attributes_doc

from acconeer.exptool import a121
from acconeer.exptool import type_migration as tm
from acconeer.exptool._core.class_creation.attrs import attrs_ndarray_isclose
from acconeer.exptool.a121._core.entities.configs.config_enums import IdleState, Profile
from acconeer.exptool.a121._core.utils import is_divisor_of, is_multiple_of
from acconeer.exptool.a121._h5_utils import _create_h5_string_dataset
from acconeer.exptool.a121.algo import (
    APPROX_BASE_STEP_LENGTH_M,
    ENVELOPE_FWHM_M,
    AlgoBase,
    AlgoConfigBase,
    Controller,
    select_prf,
)
from acconeer.exptool.a121.algo._utils import (
    estimate_frame_rate,
    get_max_profile_without_direct_leakage,
    get_max_step_length,
)

from ._processors import Processor, ProcessorConfig, ProcessorContext, ProcessorExtraResult
from ._subsweep_utils import get_subsweep_configs


SPARSE_IQ_PPC = 24


def optional_profile_converter(profile: Optional[Profile]) -> Optional[Profile]:
    if profile is None:
        return None

    return Profile(profile)


def idle_state_converter(idle_state: IdleState) -> IdleState:
    return IdleState(idle_state)


@attributes_doc
@attrs.mutable(kw_only=True)
class DetectorConfig(AlgoConfigBase):
    start_m: float = attrs.field(default=0.3)
    """Start point of measurement interval in meters."""

    end_m: float = attrs.field(default=2.5)
    """End point of measurement interval in meters."""

    profile: Optional[a121.Profile] = attrs.field(
        default=None, converter=optional_profile_converter
    )
    """Sets the profile. If no argument is provided, the highest possible
    profile without interference of direct leakage is used to maximize SNR."""

    step_length: Optional[int] = attrs.field(default=None)
    """Step length in points. If no argument is provided, the step length is automatically
    calculated based on the profile."""

    frame_rate: float = attrs.field(default=12.0)
    """Frame rate in Hz."""

    sweeps_per_frame: int = attrs.field(default=16)
    """Number of sweeps per frame."""

    automatic_subsweeps: bool = attrs.field(default=False)
    """Automatically select subsweeps and configure HWAAS."""

    signal_quality: float = attrs.field(default=15.0)
    """Signal quality measurement (higher = better signal, lower = less power consumption)."""

    hwaas: int = attrs.field(default=32)
    """Number of HWAAS."""

    inter_frame_idle_state: a121.IdleState = attrs.field(
        default=a121.IdleState.DEEP_SLEEP, converter=idle_state_converter
    )
    """Sets the inter frame idle state."""

    intra_enable: bool = attrs.field(default=True)
    """Enables the intra-frame presence detection used for detecting
    faster movements inside frames."""

    intra_detection_threshold: float = attrs.field(default=1.3)
    """Detection threshold for the intra-frame presence detection."""

    intra_frame_time_const: float = attrs.field(default=0.15)
    """Time constant for the depthwise filtering in the intra-frame part."""

    intra_output_time_const: float = attrs.field(default=0.3)
    """Time constant for the output in the intra-frame part."""

    inter_enable: bool = attrs.field(default=True)
    """Enables the inter-frame presence detection used for detecting
    slower movements between frames."""

    inter_detection_threshold: float = attrs.field(default=1)
    """Detection threshold for the inter-frame presence detection."""

    inter_frame_fast_cutoff: float = attrs.field(default=6.0)
    """Cutoff frequency of the low pass filter for the fast filtered absolute sweep mean.
    No filtering is applied if the cutoff is set over half the frame rate (Nyquist limit)."""

    inter_frame_slow_cutoff: float = attrs.field(default=0.2)
    """Cutoff frequency of the low pass filter for the slow filtered absolute sweep mean."""

    inter_frame_deviation_time_const: float = attrs.field(default=0.5)
    """Time constant of the low pass filter for the inter-frame deviation between fast and slow."""

    inter_output_time_const: float = attrs.field(default=2)
    """Time constant for the output in the inter-frame part."""

    inter_frame_presence_timeout: Optional[int] = attrs.field(default=3)
    """Number of seconds the inter-frame presence score needs to decrease before exponential
    scaling starts for faster decline."""

    @step_length.validator
    def _validate_step_length(self, _: Any, step_length: int) -> None:
        if step_length is not None and not (
            is_divisor_of(SPARSE_IQ_PPC, step_length) or is_multiple_of(SPARSE_IQ_PPC, step_length)
        ):
            msg = f"step_length must be a divisor or multiple of {SPARSE_IQ_PPC}"
            raise ValueError(msg)

    def _collect_validation_results(self) -> list[a121.ValidationResult]:
        validation_results: list[a121.ValidationResult] = []
        bad_sensor_config = False

        if self.sweeps_per_frame <= Processor.NOISE_ESTIMATION_DIFF_ORDER:
            validation_results.append(
                a121.ValidationError(
                    self,
                    "sweeps_per_frame",
                    f"Must be greater than {Processor.NOISE_ESTIMATION_DIFF_ORDER}",
                )
            )

        if self.start_m > self.end_m:
            validation_results.append(
                a121.ValidationError(
                    self, "end_m", "End point must be further or equal to start point"
                )
            )
            bad_sensor_config = True

        if self.end_m > int(a121.PRF.PRF_5_2_MHz.mmd):
            validation_results.append(
                a121.ValidationError(
                    self,
                    "end_m",
                    f"End of range is too far (max: {int(a121.PRF.PRF_5_2_MHz.mmd)} m), "
                    + "try to decrease the end point.",
                )
            )
            bad_sensor_config = True

        if bad_sensor_config:
            # Return to avoid calling get_sensor_config with bad config.
            return validation_results

        for res in Detector._get_sensor_config(self)._collect_validation_results():
            res.source = self

            if res.aspect == "num_points":
                validation_results.append(
                    a121.ValidationError(
                        self,
                        "start_m",
                        "Range is too long. Increasing the range start reduces buffer usage.",
                    )
                )
                validation_results.append(
                    a121.ValidationError(
                        self,
                        "end_m",
                        "Range is too long. Decreasing the range end reduces buffer usage.",
                    )
                )
            else:
                validation_results.append(res)

        return validation_results


@attrs.mutable(kw_only=True)
class DetectorContext(AlgoBase):
    estimated_frame_rate: float = attrs.field(default=None)


@attrs.frozen(kw_only=True)
class DetectorMetadata:
    start_m: float = attrs.field()
    """Actual start point of measurement in meters"""

    end_m: float = attrs.field()
    """Actual end point of measurement in meters"""

    step_length_m: Optional[float] = attrs.field()
    """Actual step length between each data point of the measurement in meters. (Only valid if automatic_subsweeps is False)"""

    num_points: int = attrs.field()
    """The number of data points in the measurement"""

    profile: Optional[a121.Profile] = attrs.field()
    """Profile used for measurement. (Only valid if automatic_subsweeps is False)"""


@attrs.frozen(kw_only=True)
class DetectorResult:
    intra_presence_score: float = attrs.field()
    """A measure of the amount of fast motion detected."""

    intra_depthwise_scores: npt.NDArray[np.float64] = attrs.field(eq=attrs_ndarray_isclose)
    """The depthwise presence scores for fast motions."""

    inter_presence_score: float = attrs.field()
    """A measure of the amount of slow motion detected."""

    inter_depthwise_scores: npt.NDArray[np.float64] = attrs.field(eq=attrs_ndarray_isclose)
    """The depthwise presence scores for slow motions."""

    presence_distance: float = attrs.field()
    """The distance, in meters, to the detected object."""

    presence_detected: bool = attrs.field()
    """True if presence was detected, False otherwise."""

    processor_extra_result: ProcessorExtraResult = attrs.field()
    service_result: a121.Result = attrs.field()


class Detector(Controller[DetectorConfig, DetectorResult]):
    MIN_DIST_M = {
        a121.Profile.PROFILE_1: None,
        a121.Profile.PROFILE_2: 2 * ENVELOPE_FWHM_M[a121.Profile.PROFILE_2],
        a121.Profile.PROFILE_3: 2 * ENVELOPE_FWHM_M[a121.Profile.PROFILE_3],
        a121.Profile.PROFILE_4: 2 * ENVELOPE_FWHM_M[a121.Profile.PROFILE_4],
        a121.Profile.PROFILE_5: 2 * ENVELOPE_FWHM_M[a121.Profile.PROFILE_5],
    }

    def __init__(
        self,
        *,
        client: a121.Client,
        sensor_id: int,
        detector_config: DetectorConfig,
        detector_context: Optional[DetectorContext] = None,
    ) -> None:
        super().__init__(client=client, config=detector_config)
        self.sensor_id = sensor_id
        self.detector_metadata: Optional[DetectorMetadata] = None
        self.detector_context = detector_context

        self.started = False

    def start(
        self,
        recorder: Optional[a121.Recorder] = None,
        _algo_group: Optional[h5py.Group] = None,
    ) -> None:
        if self.started:
            msg = "Already started"
            raise RuntimeError(msg)

        sensor_config = self._get_sensor_config(self.config)
        self.session_config = a121.SessionConfig(
            {self.sensor_id: sensor_config},
            extended=False,
        )

        if self.detector_context is None:
            self.estimated_frame_rate = estimate_frame_rate(self.client, self.session_config)
            self.detector_context = DetectorContext(estimated_frame_rate=self.estimated_frame_rate)
        else:
            self.estimated_frame_rate = self.detector_context.estimated_frame_rate

        # Add estimated frame rate to context if it differs more than
        # 10% from the set frame rate
        if (
            np.abs(self.config.frame_rate - self.estimated_frame_rate) / self.config.frame_rate
            > 0.1
        ):
            processor_context = ProcessorContext(estimated_frame_rate=self.estimated_frame_rate)
        else:
            processor_context = ProcessorContext(estimated_frame_rate=None)

        metadata = self.client.setup_session(self.session_config)
        assert isinstance(metadata, a121.Metadata)

        processor_config = self._get_processor_config(self.config)

        start_m = sensor_config.subsweeps[0].start_point * APPROX_BASE_STEP_LENGTH_M
        end_m = (
            sensor_config.subsweeps[-1].start_point
            + (sensor_config.subsweeps[-1].num_points - 1)
            * sensor_config.subsweeps[-1].step_length
        ) * APPROX_BASE_STEP_LENGTH_M
        step_length_m = metadata.base_step_length_m * sensor_config.subsweeps[0].step_length
        num_points = sum([subsweep.num_points for subsweep in sensor_config.subsweeps])
        profile = sensor_config.subsweeps[0].profile

        self.detector_metadata = DetectorMetadata(
            start_m=start_m,
            end_m=end_m,
            step_length_m=step_length_m if sensor_config.num_subsweeps == 1 else None,
            num_points=num_points,
            profile=profile if sensor_config.num_subsweeps == 1 else None,
        )

        self.processor = Processor(
            sensor_config=sensor_config,
            metadata=metadata,
            processor_config=processor_config,
            context=processor_context,
        )

        if recorder is not None:
            if isinstance(recorder, a121.H5Recorder):
                if _algo_group is None:
                    _algo_group = recorder.require_algo_group("presence_detector")
                _record_algo_data(
                    _algo_group,
                    self.sensor_id,
                    self.config,
                    self.detector_context,
                )
            else:
                # Should never happen as we currently only have the H5Recorder
                warnings.warn("Will not save algo data")

            self.client.attach_recorder(recorder)

        self.client.start_session()

        self.started = True

    @classmethod
    def _get_sensor_config(cls, config: DetectorConfig) -> a121.SensorConfig:
        if config.automatic_subsweeps:
            subsweep_config = get_subsweep_configs(
                config.start_m, config.end_m, config.signal_quality
            )
            sensor_config = a121.SensorConfig(
                subsweeps=subsweep_config,
                sweeps_per_frame=config.sweeps_per_frame,
                frame_rate=config.frame_rate,
                inter_frame_idle_state=config.inter_frame_idle_state,
            )
        else:
            if config.profile is not None:
                profile = config.profile
            else:
                profile = get_max_profile_without_direct_leakage(config.start_m)
            if config.step_length is not None:
                step_length = config.step_length
            else:
                step_length = get_max_step_length(profile)

            start_point = int(np.floor(config.start_m / APPROX_BASE_STEP_LENGTH_M))
            num_point = int(
                np.ceil(
                    (config.end_m - config.start_m) / (step_length * APPROX_BASE_STEP_LENGTH_M)
                )
                + 1
            )
            end_point = start_point + (num_point - 1) * step_length
            sensor_config = a121.SensorConfig(
                profile=profile,
                start_point=start_point,
                num_points=num_point,
                step_length=step_length,
                prf=select_prf(end_point, profile),
                hwaas=config.hwaas,
                sweeps_per_frame=config.sweeps_per_frame,
                frame_rate=config.frame_rate,
                inter_frame_idle_state=config.inter_frame_idle_state,
            )
        return sensor_config

    @classmethod
    def _get_processor_config(cls, config: DetectorConfig) -> ProcessorConfig:
        return ProcessorConfig(
            intra_enable=config.intra_enable,
            intra_detection_threshold=config.intra_detection_threshold,
            intra_frame_time_const=config.intra_frame_time_const,
            intra_output_time_const=config.intra_output_time_const,
            inter_enable=config.inter_enable,
            inter_detection_threshold=config.inter_detection_threshold,
            inter_frame_fast_cutoff=config.inter_frame_fast_cutoff,
            inter_frame_slow_cutoff=config.inter_frame_slow_cutoff,
            inter_frame_deviation_time_const=config.inter_frame_deviation_time_const,
            inter_output_time_const=config.inter_output_time_const,
            inter_frame_presence_timeout=config.inter_frame_presence_timeout,
        )

    def get_next(self) -> DetectorResult:
        if not self.started:
            msg = "Not started"
            raise RuntimeError(msg)

        result = self.client.get_next()
        assert isinstance(result, a121.Result)

        assert self.processor is not None
        processor_result = self.processor.process(result)

        return DetectorResult(
            intra_presence_score=processor_result.intra_presence_score,
            intra_depthwise_scores=processor_result.intra,
            inter_presence_score=processor_result.inter_presence_score,
            inter_depthwise_scores=processor_result.inter,
            presence_distance=processor_result.presence_distance,
            presence_detected=processor_result.presence_detected,
            processor_extra_result=processor_result.extra_result,
            service_result=result,
        )

    def update_config(self, config: DetectorConfig) -> None:
        raise NotImplementedError

    def stop_detector(self) -> Any:
        if not self.started:
            msg = "Already stopped"
            raise RuntimeError(msg)

        self.client.stop_session()
        self.started = False

        return None

    def stop_recorder(self) -> Any:
        recorder = self.client.detach_recorder()
        if recorder is None:
            recorder_result = None
        else:
            recorder_result = recorder.close()

        return recorder_result

    def stop(self) -> Any:
        self.stop_detector()
        recorder_result = self.stop_recorder()

        return recorder_result


def _record_algo_data(
    algo_group: h5py.Group,
    sensor_id: int,
    config: DetectorConfig,
    context: DetectorContext,
) -> None:
    algo_group.create_dataset(
        "sensor_id",
        data=sensor_id,
        track_times=False,
    )
    _create_h5_string_dataset(algo_group, "detector_config", config.to_json())
    _create_h5_string_dataset(algo_group, "detector_context", context.to_json())


def _load_algo_data(
    algo_group: h5py.Group,
) -> Tuple[int, DetectorConfig, Optional[DetectorContext]]:
    sensor_id = algo_group["sensor_id"][()]
    try:
        config = detector_config_timeline.migrate(algo_group["detector_config"][()].decode())
    except tm.core.MigrationErrorGroup as exc:
        raise TypeError() from exc

    context_data_set = algo_group.get("detector_context")
    if context_data_set is None:
        context = None
    else:
        context = DetectorContext.from_json(context_data_set[()])

    return sensor_id, config, context


@attrs.mutable(kw_only=True)
class _DetectorConfig_v0(AlgoConfigBase):
    start_m: float = attrs.field(default=0.3)
    end_m: float = attrs.field(default=2.5)
    profile: Optional[a121.Profile] = attrs.field(
        default=None, converter=optional_profile_converter
    )
    step_length: Optional[int] = attrs.field(default=None)
    frame_rate: float = attrs.field(default=12.0)
    sweeps_per_frame: int = attrs.field(default=16)
    automatic_subsweeps: bool = attrs.field(default=False)
    signal_quality: float = attrs.field(default=15.0)
    hwaas: int = attrs.field(default=32)
    inter_frame_idle_state: a121.IdleState = attrs.field(
        default=a121.IdleState.DEEP_SLEEP, converter=idle_state_converter
    )
    intra_enable: bool = attrs.field(default=True)
    intra_detection_threshold: float = attrs.field(default=1.3)
    intra_frame_time_const: float = attrs.field(default=0.15)
    intra_output_time_const: float = attrs.field(default=0.3)
    inter_enable: bool = attrs.field(default=True)
    inter_detection_threshold: float = attrs.field(default=1)
    inter_frame_fast_cutoff: float = attrs.field(default=6.0)
    inter_frame_slow_cutoff: float = attrs.field(default=0.2)
    inter_frame_deviation_time_const: float = attrs.field(default=0.5)
    inter_output_time_const: float = attrs.field(default=2)
    inter_frame_presence_timeout: Optional[int] = attrs.field(default=3)
    inter_phase_boost: bool = attrs.field(default=False)

    def _collect_validation_results(self) -> list[a121.ValidationResult]:
        return []

    def migrate(self) -> DetectorConfig:
        return DetectorConfig(
            start_m=self.start_m,
            end_m=self.end_m,
            profile=self.profile,
            step_length=self.step_length,
            frame_rate=self.frame_rate,
            sweeps_per_frame=self.sweeps_per_frame,
            automatic_subsweeps=self.automatic_subsweeps,
            signal_quality=self.signal_quality,
            hwaas=self.hwaas,
            inter_frame_idle_state=self.inter_frame_idle_state,
            intra_enable=self.intra_enable,
            intra_detection_threshold=self.intra_detection_threshold,
            intra_frame_time_const=self.intra_frame_time_const,
            intra_output_time_const=self.intra_output_time_const,
            inter_enable=self.inter_enable,
            inter_detection_threshold=self.inter_detection_threshold,
            inter_frame_fast_cutoff=self.inter_frame_fast_cutoff,
            inter_frame_slow_cutoff=self.inter_frame_slow_cutoff,
            inter_frame_deviation_time_const=self.inter_frame_deviation_time_const,
            inter_output_time_const=self.inter_output_time_const,
            inter_frame_presence_timeout=self.inter_frame_presence_timeout,
            # inter_phase_boost is removed
        )


detector_config_timeline = (
    tm.start(_DetectorConfig_v0)
    .load(str, _DetectorConfig_v0.from_json, fail=[TypeError])
    .load(dict, _DetectorConfig_v0.from_dict, fail=[TypeError])
    .nop()
    .epoch(DetectorConfig, _DetectorConfig_v0.migrate, fail=[])
    .load(str, DetectorConfig.from_json, fail=[TypeError])
    .load(dict, DetectorConfig.from_dict, fail=[TypeError])
)



# Copyright (c) Acconeer AB, 2023-2025
# All rights reserved
"""
API for generating power regions (for more info see ./domain.py).

The different functions return regions based on the figures below:

    ^
    |    ___           ___
    |    |X|           |X|
    |    |X|           |X|
    | ---|X|-----------|X|----
    +-------------------------->

          ^ |_________| ^
          3      2      3
    |______________________ ...
                 1

1. Session (indefinitely repeats 2. and 3.)
2. Group idle (Idle/Hibernate/Off)
3. Group active is described by the figure below

    ^
    |      ___     ___
    |      |X|-----|X|     _____
    |    __|X|XXXXX|X|__   |XXX|
    | ---|X|X|XXXXX|X|X|---|XXX|---
    +---------------------------->

          ^ ^ |___| ^ ^    |___|
          4 2   3   2 4      5*
         |_____________________|
                  1
1. Group active
2. Frame active for every SensorConfig in the SessionConfig
3. Frame idle. Idles with the previous (chronologically)
               SensorConfig's inter_frame_idle_state
4. Going to/returning from Idle/Hibernate/Off.
5. Algorithm Processing.
*. Different algorithms (passed via the "algorithm"-parameter in most functions)
   are free to decide the ordering of the regions in Group active
   (including addition of processing regions like 4). This is needed to model
   different algorithms' control flow. (See ./algo.py)

Going deeper down into, the frame active consists of sweep active and sweep idle, etc.
"""

from __future__ import annotations

import collections
import typing as t

from acconeer.exptool import a121
from acconeer.exptool.a121._core import utils as core_utils

from . import algo, domain
from . import dependency_injection as di
from .lookup import Module, Sensor


_T = t.TypeVar("_T")

_ms = _mA = 1e-3
_us = _uA = 1e-6


def configured_rate(config: a121.SessionConfig) -> t.Optional[float]:
    """
    Returns the configured rate of the session config.
    Either returns the update_rate or a frame_rate.
    """
    if config.update_rate is not None:
        return config.update_rate

    frame_rates = set(
        sensor_config.frame_rate
        for sensor_config in core_utils.iterate_extended_structure_values(config.groups)
        if sensor_config.frame_rate is not None
    )

    if len(frame_rates) == 1:
        (frame_rate,) = frame_rates
        return frame_rate
    else:
        return None


def is_high_speed_mode(config: a121.SensorConfig) -> bool:
    hsm_profiles = {a121.Profile.PROFILE_3, a121.Profile.PROFILE_4, a121.Profile.PROFILE_5}
    return (
        config.num_subsweeps == 1
        and config.profile in hsm_profiles
        and config.inter_sweep_idle_state == a121.IdleState.READY
        and not config.continuous_sweep_mode
    )


def power_state(
    state: Sensor.LowerIdleState,
    duration: float,
    sensor: Sensor = di.DEFAULT_SENSOR,
    module: Module = di.DEFAULT_MODULE,
) -> domain.SimpleRegion:
    """
    Returns a simple region that describes being in the given power state
    """
    return domain.SimpleRegion(
        (sensor.inter_frame_currents[state] + module.currents[Module.PowerState.ASLEEP]),
        duration,
        description=state.name,
        tag=domain.EnergyRegion.Tag.IDLE,
    )


def subsweep_active(
    config: a121.SubsweepConfig,
    high_speed_mode: bool,
    subsweep_index: int,
    sensor: Sensor = di.DEFAULT_SENSOR,
    module: Module = di.DEFAULT_MODULE,
) -> domain.SimpleRegion:
    """
    Returns a region that describes measuring a single subsweep (no overheads)
    """
    if high_speed_mode:
        overhead_duration = sensor.high_speed_mode_overheads[Sensor.FixedOverhead.SUBSWEEP]
    else:
        overhead_duration = sensor.normal_overheads[Sensor.FixedOverhead.SUBSWEEP]

    sample_duration = sensor.sample_durations[config.profile][config.prf]
    point_duration = (
        config.hwaas * sample_duration + sensor.point_overheads[config.profile][config.prf]
    )
    subsweep_duration = config.num_points * point_duration + (
        3 * sample_duration + overhead_duration
    )

    return domain.SimpleRegion(
        (sensor.measure_currents[config.profile] + module.currents[Module.PowerState.AWAKE]),
        subsweep_duration,
        domain.EnergyRegion.Tag.MEASURE,
        description=f"Subsweep {subsweep_index} measure ({config.profile.name}, {config.num_points} points)",
    )


def sweep_active(
    config: a121.SensorConfig,
    sensor: Sensor = di.DEFAULT_SENSOR,
    module: Module = di.DEFAULT_MODULE,
) -> domain.EnergyRegion:
    """
    Returns a region that describes measuring a sweep, subsweeps and subsweep overhead included
    """
    high_speed_mode = is_high_speed_mode(config)
    if high_speed_mode:
        overhead_duration = sensor.high_speed_mode_overheads[Sensor.FixedOverhead.SWEEP]
    else:
        overhead_duration = sensor.normal_overheads[Sensor.FixedOverhead.SWEEP]

    overhead = domain.SimpleRegion(
        module.currents[Module.PowerState.AWAKE],
        duration=overhead_duration,
        description="inter-subsweep overhead",
        tag=domain.EnergyRegion.Tag.OVERHEAD,
    )

    return overhead.join(
        (
            subsweep_active(subsweep, high_speed_mode, idx, sensor, module)
            for idx, subsweep in enumerate(config.subsweeps, start=1)
        ),
        description="Sweep active",
    )


def sweep_idle(
    config: a121.SensorConfig,
    duration: float,
    sensor: Sensor = di.DEFAULT_SENSOR,
    module: Module = di.DEFAULT_MODULE,
) -> domain.SimpleRegion:
    """
    Returns a region that describes the idling between sweeps
    """
    current = module.currents[Module.PowerState.AWAKE]

    if is_high_speed_mode(config):
        current += sensor.inter_sweep_currents_high_speed_mode[config.inter_sweep_idle_state]
        description = f"Sweep idle ({config.inter_sweep_idle_state.name}, High speed mode)"
    else:
        current += sensor.inter_sweep_currents[config.inter_sweep_idle_state]
        description = f"Sweep idle ({config.inter_sweep_idle_state.name})"

    return domain.SimpleRegion(
        current,
        duration,
        tag=domain.EnergyRegion.Tag.IDLE,
        description=description,
    )


def frame_active(
    config: a121.SensorConfig,
    sensor: Sensor = di.DEFAULT_SENSOR,
    module: Module = di.DEFAULT_MODULE,
) -> domain.CompositeRegion:
    """
    Returns a region that describes the active part of a frame (sweeps and sweep idles)
    """
    if is_high_speed_mode(config):
        overhead_duration = sensor.high_speed_mode_overheads[Sensor.FixedOverhead.FRAME]
    else:
        overhead_duration = sensor.normal_overheads[Sensor.FixedOverhead.FRAME]

    frame_overhead = domain.SimpleRegion(
        module.currents[Module.PowerState.AWAKE],  # FIXME: guessed
        overhead_duration,
        tag=domain.EnergyRegion.Tag.OVERHEAD,
        description="Frame overhead",
    )

    arbitrary_setup = domain.SimpleRegion(
        (
            # looks to be about this
            module.currents[Module.PowerState.AWAKE]
            + sensor.inter_frame_currents[config.inter_frame_idle_state]
        ),
        sensor.pre_measure_setup_duration,
        tag=domain.EnergyRegion.Tag.COMMUNICATION,
        description="Setup",
    )

    read = domain.SimpleRegion(
        (
            (module.currents[Module.PowerState.IO] + module.currents[Module.PowerState.AWAKE]) / 2
            + sensor.inter_frame_currents[config.inter_frame_idle_state]
        ),
        sensor.read_duration,
        tag=domain.EnergyRegion.Tag.COMMUNICATION,
        description="Read",
    )

    active = sweep_active(config, sensor, module)
    minimum_transition_time = sensor.time_for_measure_transition[config.inter_sweep_idle_state]

    if config.sweep_rate is None:
        inter_sweep_idle_duration = minimum_transition_time
    else:
        inter_sweep_idle_duration = max(
            minimum_transition_time, 1 / config.sweep_rate - active.duration
        )

    return domain.CompositeRegion(
        (
            arbitrary_setup,
            frame_overhead,
            sweep_idle(
                config,
                inter_sweep_idle_duration,
                sensor,
                module,
            ).join([active] * config.sweeps_per_frame, description="Frame measurement"),
            read,
        ),
        description="Frame active",
    )


def frame_idle(
    inter_frame_idle_state: a121.IdleState,
    duration: float,
    sensor: Sensor = di.DEFAULT_SENSOR,
    module: Module = di.DEFAULT_MODULE,
) -> domain.SimpleRegion:
    """
    Returns a region that describes the idling between frames
    """
    return domain.SimpleRegion(
        (
            sensor.inter_frame_currents[inter_frame_idle_state]
            + module.currents[Module.PowerState.AWAKE]
        ),
        duration,
        tag=domain.EnergyRegion.Tag.IDLE,
        description=f"Frame idle ({inter_frame_idle_state.name})",
    )


def group_active(
    session_config: a121.SessionConfig,
    lower_idle_state: t.Optional[Sensor.LowerIdleState],
    algorithm: algo.Algorithm = di.DEFAULT_ALGO,
    sensor: Sensor = di.DEFAULT_SENSOR,
    module: Module = di.DEFAULT_MODULE,
) -> domain.CompositeRegion:
    """
    Describes the active part of a group.
    """
    frame_actives = [
        {sid: frame_active(sensor_config, sensor, module) for sid, sensor_config in group.items()}
        for group in session_config.groups
    ]

    frame_idles = [
        {
            sid: frame_idle(
                sensor_config.inter_frame_idle_state,
                sensor.time_for_measure_transition[sensor_config.inter_frame_idle_state],
                sensor,
                module,
            )
            for sid, sensor_config in group.items()
        }
        for group in session_config.groups
    ]

    return domain.CompositeRegion(
        tuple(
            algorithm.decide_control(
                frame_actives,
                frame_idles,
                session_config,
                lower_idle_state,
                sensor,
                module,
            )
        ),
        description="Measure groups active",
    )


def group_idle(
    inter_group_power_state: t.Union[a121.IdleState, Sensor.LowerIdleState],
    duration: float,
    sensor: Sensor = di.DEFAULT_SENSOR,
    module: Module = di.DEFAULT_MODULE,
) -> domain.SimpleRegion:
    """
    Returns a region that describes the idling between groups
    """
    if inter_group_power_state is None:
        module_power_state = Module.PowerState.AWAKE
    else:
        module_power_state = Module.PowerState.ASLEEP

    return domain.SimpleRegion(
        (
            sensor.inter_frame_currents[inter_group_power_state]
            + module.currents[module_power_state]
        ),
        duration,
        tag=domain.EnergyRegion.Tag.IDLE,
        description=f"Group idle ({inter_group_power_state.name})",
    )


def session_generator(
    session_config: a121.SessionConfig,
    lower_idle_state: t.Optional[Sensor.LowerIdleState],
    algorithm: algo.Algorithm = di.DEFAULT_ALGO,
    sensor: Sensor = di.DEFAULT_SENSOR,
    module: Module = di.DEFAULT_MODULE,
) -> t.Iterator[domain.EnergyRegion]:
    """
    Indefinitely simulates the session, yielding region per region
    """
    active = group_active(session_config, lower_idle_state, algorithm, sensor, module)

    rate = configured_rate(session_config)

    regions: list[domain.EnergyRegion]
    if rate is None:
        regions = [active]
    else:
        duration = 1 / rate - active.duration

        if duration > 0:
            idle = group_idle(
                lower_idle_state or algo.last_inter_frame_idle_state(session_config),
                duration,
                sensor=sensor,
                module=module,
            )
            regions = [active, idle]
        else:
            regions = [active]

    while True:
        yield from regions


def session(
    session_config: a121.SessionConfig,
    lower_idle_state: t.Optional[Sensor.LowerIdleState],
    duration: t.Optional[float] = None,
    num_actives: t.Optional[int] = None,
    algorithm: algo.Algorithm = di.DEFAULT_ALGO,
    sensor: Sensor = di.DEFAULT_SENSOR,
    module: Module = di.DEFAULT_MODULE,
) -> domain.CompositeRegion:
    """
    Models the first `duration` seconds of a session
    """
    regions: list[domain.EnergyRegion] = []
    elapsed_time = 0.0
    found_actives = 0

    if num_actives is None and duration is None:
        msg = "At least one of 'num_actives' and 'duration' needs to be not None"
        raise ValueError(msg)

    for region in session_generator(
        session_config,
        lower_idle_state,
        algorithm,
        sensor,
        module,
    ):
        regions += [region]
        elapsed_time += region.duration

        if not (
            isinstance(region, domain.SimpleRegion) and region.tag is domain.EnergyRegion.Tag.IDLE
        ):
            found_actives += 1

        if duration is not None and elapsed_time > duration:
            break
        if num_actives is not None and found_actives >= num_actives:
            break

    sequence = domain.CompositeRegion(
        tuple(regions),
        f"First {duration} seconds of session",
    )

    if duration is not None:
        return sequence.truncate(duration)
    else:
        return sequence


def converged_average_current(
    session_config: a121.SessionConfig,
    lower_idle_state: t.Optional[Sensor.LowerIdleState],
    absolute_tolerance: float,
    convergence_window: int = 10,
    algorithm: algo.Algorithm = di.DEFAULT_ALGO,
    sensor: Sensor = di.DEFAULT_SENSOR,
    module: Module = di.DEFAULT_MODULE,
) -> float:
    """
    Simulates the session until the average current has been within 'absolute_tolerance'.
    for 'convergence_window' iterations.
    """
    gen = session_generator(
        session_config,
        lower_idle_state,
        algorithm,
        sensor,
        module,
    )

    first = next(gen)
    total_duration = first.duration
    cumulative_charge = first.charge
    average_currents = collections.deque([first.average_current], maxlen=convergence_window)

    for region in gen:
        cumulative_charge += region.charge
        total_duration += region.duration

        average_currents.append(cumulative_charge / total_duration)
        if len(average_currents) == convergence_window and all(
            ac - min(average_currents) < absolute_tolerance for ac in average_currents
        ):
            break

    return average_currents[-1]


def dump_region(region: domain.EnergyRegion, indent: str = "") -> None:
    if isinstance(region, domain.SimpleRegion):
        print(
            "\n".join(
                [
                    f"{indent}{region.description}",
                    f"{indent} - current:  {region.average_current * 1000:.3f} mA",
                    f"{indent} - duration: {region.duration * 1000:.3f} ms",
                ]
            )
        )
    elif isinstance(region, domain.CompositeRegion):
        print(
            f"{indent}{region.description: <{50 - len(indent)}} (avg. {region.average_current * 1000:.3f} mA, {region.duration * 1000:.3f} ms)"
        )
        for r in region.regions:
            dump_region(r, indent=indent + "|   ")



# Copyright (c) Acconeer AB, 2022
# All rights reserved

import numpy as np

from PySide6 import QtCore

import pyqtgraph as pg

import acconeer.exptool as et

from .constants import HISTORY_LENGTH_S


OUTPUT_MAX_SIGNAL = 20000
OUTPUT_MAX_REL_DEV = 0.5
DETECTION_SHOW_S = 2


class PGUpdater:
    def __init__(self, sensor_config, processing_config, session_info):
        self.sensor_config = sensor_config
        self.processing_config = processing_config

        assert sensor_config.update_rate is not None

        self.setup_is_done = False

    def setup(self, win):
        win.setWindowTitle("Acconeer Button Press Example")

        self.limit_lines = []
        dashed_pen = pg.mkPen("k", width=2.5, style=QtCore.Qt.DashLine)

        self.sign_hist_plot = win.addPlot(title="Signal history")
        self.sign_hist_plot.setMenuEnabled(False)
        self.sign_hist_plot.setMouseEnabled(x=False, y=False)
        self.sign_hist_plot.hideButtons()
        self.sign_hist_plot.addLegend()
        self.sign_hist_plot.showGrid(x=True, y=True)
        self.sign_hist_plot.setLabel("bottom", "Time (s)")
        self.sign_hist_plot.setXRange(-HISTORY_LENGTH_S, 0)
        self.sign_hist_plot.setYRange(0, OUTPUT_MAX_SIGNAL)
        self.sign_hist_curve = self.sign_hist_plot.plot(
            pen=et.utils.pg_pen_cycler(0),
            name="Envelope signal",
        )
        self.sign_lp_hist_curve = self.sign_hist_plot.plot(
            pen=et.utils.pg_pen_cycler(1),
            name="Filtered envelope signal",
        )

        win.nextRow()

        self.rel_dev_hist_plot = win.addPlot(title="Relative deviation history")
        self.rel_dev_hist_plot.setMenuEnabled(False)
        self.rel_dev_hist_plot.setMouseEnabled(x=False, y=False)
        self.rel_dev_hist_plot.hideButtons()
        self.rel_dev_hist_plot.showGrid(x=True, y=True)
        self.rel_dev_hist_plot.setLabel("bottom", "Time (s)")
        self.rel_dev_hist_plot.setXRange(-HISTORY_LENGTH_S, 0)
        self.rel_dev_hist_plot.setYRange(0, OUTPUT_MAX_REL_DEV)
        self.rel_dev_lp_hist_curve = self.rel_dev_hist_plot.plot(
            pen=et.utils.pg_pen_cycler(0),
            name="Relative deviation",
        )

        self.detection_dots = self.rel_dev_hist_plot.plot(
            pen=None,
            symbol="o",
            symbolSize=20,
            symbolBrush=et.utils.color_cycler(1),
            name="Detections",
        )

        limit_line = pg.InfiniteLine(angle=0, pen=dashed_pen)
        self.rel_dev_hist_plot.addItem(limit_line)

        self.limit_lines.append(limit_line)

        self.detection_html_format = (
            '<div style="text-align: center">'
            '<span style="color: #FFFFFF;font-size:16pt;">'
            "{}</span></div>"
        )
        detection_html = self.detection_html_format.format("Button press detected!")

        self.detection_text_item = pg.TextItem(
            html=detection_html,
            fill=pg.mkColor(255, 140, 0),
            anchor=(0.5, 0),
        )

        self.detection_text_item.setPos(-HISTORY_LENGTH_S / 2, 0.95 * OUTPUT_MAX_REL_DEV)
        self.rel_dev_hist_plot.addItem(self.detection_text_item)
        self.detection_text_item.hide()

        self.smooth_max_signal = et.utils.SmoothMax(
            self.sensor_config.update_rate,
            hysteresis=0.6,
            tau_decay=3,
        )

        self.smooth_max_rel_dev = et.utils.SmoothMax(
            self.sensor_config.update_rate,
            hysteresis=0.6,
            tau_decay=3,
        )

        self.setup_is_done = True
        self.update_processing_config()

    def update_processing_config(self, processing_config=None):
        if processing_config is None:
            processing_config = self.processing_config
        else:
            self.processing_config = processing_config

        if not self.setup_is_done:
            return

        self.sign_hist_plot.setVisible(True)
        self.rel_dev_hist_plot.setVisible(True)

        for line in self.limit_lines:
            line.setPos(processing_config.threshold)

    def update(self, data):
        signal_hist_ys = data["signal_history"]
        signal_lp_hist_ys = data["signal_lp_history"]
        rel_dev_lp_hist_ys = data["rel_dev_lp_history"]
        t_detections = data["detection_history"]

        hist_xs = np.linspace(-HISTORY_LENGTH_S, 0, len(signal_hist_ys))

        self.sign_hist_curve.setData(hist_xs, signal_hist_ys)
        self.sign_lp_hist_curve.setData(hist_xs, signal_lp_hist_ys)
        self.rel_dev_lp_hist_curve.setData(hist_xs, rel_dev_lp_hist_ys)
        self.detection_dots.setData(t_detections, data["threshold"] * np.ones(len(t_detections)))

        m = np.max(signal_hist_ys) if signal_hist_ys.size > 0 else 1
        self.sign_hist_plot.setYRange(0, self.smooth_max_signal.update(m))

        m = np.max(rel_dev_lp_hist_ys) if rel_dev_lp_hist_ys.size > 0 else 1e-3
        m = max(2 * data["threshold"], m)
        ymax = self.smooth_max_rel_dev.update(m)
        self.rel_dev_hist_plot.setYRange(0, ymax)
        self.detection_text_item.setPos(-HISTORY_LENGTH_S / 2, 0.95 * ymax)

        show_detection_text = t_detections.size > 0 and (-t_detections[-1]) < DETECTION_SHOW_S
        self.detection_text_item.setVisible(bool(show_detection_text))



# Copyright (c) Acconeer AB, 2023-2024
# All rights reserved
from __future__ import annotations

import json
import time
import typing as t

from .communication_protocol import CommunicationProtocol, Message
from .links import BufferedLink


_MessageT = t.TypeVar("_MessageT", bound=Message)


class MessageStreamError(Exception):
    pass


class MessageStream:
    """
    Helper object that automatically handles message parsing.

    This class takes no responsibility of the passed link.
    """

    def __init__(
        self,
        link: BufferedLink,
        protocol: type[CommunicationProtocol[t.Any]],
        message_handler: t.Callable[[Message], t.Any],
        link_error_callback: t.Callable[[Exception], t.Any],
    ) -> None:
        self._link = link
        self._error_callback = link_error_callback
        self._message_handler = message_handler

        self.protocol = protocol

        self._stream = self._get_stream()

    def send_command(self, command: bytes) -> None:
        try:
            self._link.send(command)
        except Exception as e:
            self._error_callback(e)

    def wait_for_message(
        self,
        message_type: type[_MessageT],
        timeout_s: t.Optional[float] = None,
    ) -> _MessageT:
        """Retrieves and applies messages until a message of type ``message_type`` is encountered.

        :param message_type: a subclass of ``Message``
        :param message_handler: Function that handles messages received that aren't of ``message_type``
        :param timeout_s: Limit the time spent in this function
        :raises MessageStreamError:
            if timeout_s is set and that amount of time has elapsed
            without predicate evaluating to True
        """
        deadline = None if (timeout_s is None) else time.monotonic() + timeout_s

        for message in self._stream:
            if type(message) is message_type:
                return message
            else:
                self._message_handler(message)

            if deadline is not None and time.monotonic() > deadline:
                msg = f"Deadline was reached without finding message of type {message_type.__name__!r}"
                raise MessageStreamError(msg)

        msg = "No messages to consume"
        raise MessageStreamError(msg)

    def _get_stream(self) -> t.Iterator[Message]:
        """returns an iterator of parsed messages"""
        while True:
            try:
                header_in_bytes = self._link.recv_until(self.protocol.end_sequence)
            except Exception as e:
                self._error_callback(e)

            try:
                header: dict[str, t.Any] = json.loads(header_in_bytes)
            except json.JSONDecodeError:
                self._error_callback(RuntimeError(f"Cannot decode header {header_in_bytes!r}"))

            try:
                payload_size = header["payload_size"]
            except KeyError:
                payload = bytes()
            else:
                try:
                    payload = self._link.recv(payload_size)
                except Exception as e:
                    self._error_callback(e)

            resp = self.protocol.parse_message(header, payload)
            yield resp



# Copyright (c) Acconeer AB, 2022-2025
# All rights reserved

from __future__ import annotations

import abc
from enum import Enum
from typing import (
    Any,
    Callable,
    Collection,
    Generic,
    Iterable,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

import attrs
import numpy as np
import typing_extensions as te

from PySide6 import QtCore, QtGui
from PySide6.QtWidgets import (
    QCheckBox,
    QComboBox,
    QDoubleSpinBox,
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QLayout,
    QLineEdit,
    QSlider,
    QSpinBox,
    QStackedWidget,
    QWidget,
)

from acconeer.exptool._core.entities.validation_result import Criticality
from acconeer.exptool.app.new.ui import icons
from acconeer.exptool.app.new.ui.components.data_editor import DataEditor

from .common import MaybeIterable, as_sequence


def widget_wrap_layout(layout: QLayout) -> QWidget:
    dummy = QWidget()
    dummy.setLayout(layout)
    return dummy


PidgetHook = Callable[["Pidget", Mapping[str, "Pidget"]], None]
T = TypeVar("T")
EnumT = TypeVar("EnumT", bound=Enum)


def _hooks_converter(a: MaybeIterable[PidgetHook]) -> Sequence[PidgetHook]:
    return as_sequence(a)


@attrs.frozen(kw_only=True, slots=False)
class PidgetFactory(abc.ABC):
    name_label_text: str = attrs.field()
    name_label_tooltip: Optional[str] = None
    note_label_text: Optional[str] = None
    extra_widget_factory: Callable[[], Optional[QWidget]] = lambda: None
    hooks: Sequence[PidgetHook] = attrs.field(factory=tuple, converter=_hooks_converter)

    @name_label_text.validator
    def check_for_whitespaces(self, attribute: Any, value: str) -> None:
        if value != value.strip():
            msg = "Labels cannot start or end with a whitespace"
            raise ValueError(msg)

    @name_label_text.validator
    def check_label_text_format(self, attribute: Any, value: str) -> None:
        if len(value) > 0 and value[-1] != ":":
            msg = "Labels have to end with ':'"
            raise ValueError(msg)

    @abc.abstractmethod
    def create(self, parent: QWidget) -> Pidget: ...

    def create_name_label(self, parent: QWidget) -> QLabel:
        label = QLabel(self.name_label_text, parent=parent)
        if self.name_label_tooltip is not None:
            label.setToolTip(self.name_label_tooltip)
        return label

    def create_note_label(self, parent: QWidget) -> QLabel:
        label = QLabel(parent=parent)
        label.setWordWrap(True)
        label.setContentsMargins(5, 5, 5, 5)
        return label


class Pidget(DataEditor[Any]):
    """Base class for a parameter-bound widget.

    A ``Pidget`` comes with a
    ``name`` label and an ``note`` label by default.
    """

    sig_update = QtCore.Signal(object)

    def __init__(self, factory: PidgetFactory, parent: QWidget) -> None:
        super().__init__(parent=parent)
        self.__note_widget = factory.create_note_label(self)

    def set_standard_layout(
        self,
        factory: PidgetFactory,
        *,
        first_row_elements: Iterable[
            Union[
                QWidget,
                te.Literal["name_label", "extra_widget"],
            ]
        ],
        full_row_widgets: Collection[QWidget] = (),
        colstretch: tuple[int, ...] = (),
    ) -> None:
        """
        Arranges the element according to a "standard layout"

        +---------------+-----+-----+-----+
        |               |     |     |     |
        |  first_row[0] | fr1 | fr2 | ... |
        |               |     |     |     |
        |---------------+-----+-----+-----+
        |                                 |
        |            Note label           |
        |                                 |
        +---------------------------------+

        And if ``full_row_widgets`` is specified:

        +---------------+-----+-----+-----+
        |               |     |     |     |
        |  first_row[0] | fr1 | fr2 | ... |
        |               |     |     |     |
        +---------------+-----+-----+-----+
        |         Full row widget[0]      |
        +---------------------------------+
        |         Full row widget[1]      |
        +---------------------------------+
        |                ...              |
        +---------------------------------+
        |                                 |
        |            Note label           |
        |                                 |
        +---------------------------------+

        Special strings can be elements in the ``first_row_widgets``. The strings are
        replaced with a widget internally.
        """
        layout = QGridLayout()
        layout.setContentsMargins(0, 0, 0, 0)

        self.set_note_text(factory.note_label_text)

        num_name_labels = sum(e == "name_label" for e in first_row_elements)
        if num_name_labels > 1:
            msg = "At most 1 'name_label' should be specified."
            raise ValueError(msg)

        num_extra_widget = sum(e == "extra_widget" for e in first_row_elements)
        if num_extra_widget > 1:
            msg = "At most 1 'extra_widget' should be specified."
            raise ValueError(msg)

        for col, elem in enumerate(first_row_elements):
            if elem == "name_label":
                name_label = factory.create_name_label(self)
                layout.addWidget(name_label, 0, col)
            elif elem == "extra_widget":
                if (extra_widget := factory.extra_widget_factory()) is not None:
                    layout.addWidget(extra_widget, 0, col, QtCore.Qt.AlignmentFlag.AlignLeft)
            else:
                layout.addWidget(elem, 0, col)

        for column, stretch in enumerate(colstretch):
            layout.setColumnStretch(column, stretch)

        for row, row_widget in enumerate(full_row_widgets, start=1):
            layout.addWidget(row_widget, row, 0, 1, -1)

        layout.addWidget(self.__note_widget, 1 + len(full_row_widgets), 0, 1, -1)

        self.setLayout(layout)

    def set_note_text(
        self, message: Optional[str], criticality: Optional[Criticality] = None
    ) -> None:
        if not message:
            self.__note_widget.hide()
            return

        COLOR_MAP = {
            Criticality.ERROR: icons.ERROR_RED,
            Criticality.WARNING: icons.WARNING_YELLOW,
            None: "white",
        }

        self.__note_widget.show()
        self.__note_widget.setText(message)
        self.__note_widget.setStyleSheet(
            f"background-color: {COLOR_MAP[criticality]}; color: white; font: bold italic;"
        )

    def setEnabled(self, enabled: bool) -> None:
        super(DataEditor, self).setEnabled(enabled)

    @abc.abstractmethod
    def set_data(self, value: Any) -> None:
        pass

    @abc.abstractmethod
    def get_data(self) -> Any:
        pass

    @property
    def is_ready(self) -> bool:
        return not self.__note_widget.isVisible()


@attrs.frozen(kw_only=True, slots=False)
class IntPidgetFactory(PidgetFactory):
    limits: Optional[Tuple[Optional[int], Optional[int]]] = None
    suffix: Optional[str] = None

    def create(self, parent: QWidget) -> IntPidget:
        return IntPidget(self, parent)


class IntPidget(Pidget):
    def __init__(self, factory: IntPidgetFactory, parent: QWidget) -> None:
        super().__init__(factory, parent)

        self._spin_box = _PidgetSpinBox(
            self,
            limits=factory.limits,
            suffix=factory.suffix,
        )
        self._spin_box.valueChanged.connect(self.__on_changed)

        self.set_standard_layout(
            factory,
            first_row_elements=["name_label", "extra_widget", self._spin_box],
        )

    def set_data(self, value: Any) -> None:
        assert isinstance(value, int)

        with QtCore.QSignalBlocker(self):
            self._spin_box.setValue(value)

    def get_data(self) -> int:
        return int(self._spin_box.value())

    def __on_changed(self) -> None:
        self.sig_update.emit(self._spin_box.value())


@attrs.frozen(kw_only=True, slots=False)
class StrPidgetFactory(PidgetFactory):
    def create(self, parent: QWidget) -> StrPidget:
        return StrPidget(self, parent)


class StrPidget(Pidget):
    def __init__(self, factory: StrPidgetFactory, parent: QWidget) -> None:
        super().__init__(factory, parent)

        self._line_edit = QLineEdit(self)
        self._line_edit.setMaximumWidth(WIDGET_WIDTH)
        self._line_edit.editingFinished.connect(self.__on_changed)

        self.set_standard_layout(
            factory,
            first_row_elements=["name_label", "extra_widget", self._line_edit],
        )

    def set_data(self, value: Any) -> None:
        assert isinstance(value, str)

        with QtCore.QSignalBlocker(self):
            self._line_edit.setText(value)

    def get_data(self) -> str:
        return str(self._line_edit.text())

    def __on_changed(self) -> None:
        self.sig_update.emit(self.get_data())


@attrs.frozen(kw_only=True, slots=False)
class FloatPidgetFactory(PidgetFactory):
    limits: Optional[Tuple[Optional[float], Optional[float]]] = None
    suffix: Optional[str] = None
    decimals: int = 1

    def create(self, parent: QWidget) -> FloatPidget:
        return FloatPidget(self, parent)


class FloatPidget(Pidget):
    def __init__(self, factory: FloatPidgetFactory, parent: QWidget) -> None:
        super().__init__(factory, parent)

        self._spin_box = _PidgetDoubleSpinBox(
            self,
            limits=factory.limits,
            suffix=factory.suffix,
            decimals=factory.decimals,
        )
        self._spin_box.valueChanged.connect(self.__on_changed)

        self.set_standard_layout(
            factory,
            first_row_elements=["name_label", "extra_widget", self._spin_box],
        )

    def set_data(self, value: Any) -> None:
        assert isinstance(value, (int, float))

        with QtCore.QSignalBlocker(self):
            self._spin_box.setValue(value)

    def get_data(self) -> float:
        return float(self._spin_box.value())

    def __on_changed(self) -> None:
        self.sig_update.emit(self._spin_box.value())


@attrs.frozen(kw_only=True, slots=False)
class FloatSliderPidgetFactory(FloatPidgetFactory):
    limits: Tuple[float, float]
    log_scale: bool = False
    show_limit_values: bool = True
    limit_texts: Optional[Tuple[Optional[str], Optional[str]]] = None

    def __attrs_post_init__(self) -> None:
        if self.log_scale and self.limits[0] <= 0:
            msg = "Lower limit must be > 0 when using log scale"
            raise ValueError(msg)

    def create(self, parent: QWidget) -> FloatSliderPidget:  # type: ignore[override]
        return FloatSliderPidget(self, parent)


class FloatSliderPidget(Pidget):
    def __init__(self, factory: FloatSliderPidgetFactory, parent: QWidget) -> None:
        super().__init__(factory, parent)

        self.__spin_box = _PidgetDoubleSpinBox(
            self,
            limits=factory.limits,
            suffix=factory.suffix,
            decimals=factory.decimals,
        )
        self.__spin_box.valueChanged.connect(self.__on_spin_box_changed)

        slider_widget = QWidget(self)
        slider_widget_layout = QHBoxLayout(slider_widget)
        slider_widget_layout.setContentsMargins(11, 6, 11, 0)

        lower_limit, upper_limit = factory.limits
        if factory.show_limit_values:
            slider_widget_layout.addWidget(QLabel(str(lower_limit), slider_widget))

        self._slider = _PidgetFloatSlider(
            slider_widget,
            limits=factory.limits,
            decimals=factory.decimals,
            log_scale=factory.log_scale,
        )
        self._slider.wrapped_value_changed.connect(self.__on_slider_changed)
        slider_widget_layout.addWidget(self._slider)

        if factory.show_limit_values:
            slider_widget_layout.addWidget(QLabel(str(upper_limit), slider_widget))

        if factory.limit_texts is not None:
            label_text_widget = QWidget(self)
            label_text_widget_layout = QHBoxLayout(label_text_widget)
            label_text_widget.setLayout(label_text_widget_layout)

            label_text_widget_layout.setContentsMargins(11, 0, 11, 0)
            label_text_widget_layout.setSpacing(0)

            left, right = factory.limit_texts

            if left is not None:
                label = QLabel(left, label_text_widget)
                label_text_widget_layout.addWidget(label)

            label_text_widget_layout.addStretch(1)

            if right is not None:
                label = QLabel(right, label_text_widget)
                label_text_widget_layout.addWidget(label)

            full_row_widgets = [slider_widget, label_text_widget]
        else:
            full_row_widgets = [slider_widget]

        slider_widget.setLayout(slider_widget_layout)

        self.set_standard_layout(
            factory,
            first_row_elements=["name_label", "extra_widget", self.__spin_box],
            full_row_widgets=full_row_widgets,
        )

    def set_data(self, value: Any) -> None:
        assert isinstance(value, (int, float))

        with QtCore.QSignalBlocker(self):
            self.__spin_box.setValue(value)
            self._slider.wrapped_set_value(value)

    def get_data(self) -> float:
        return float(self.__spin_box.value())

    def __on_spin_box_changed(self, value: float) -> None:
        with QtCore.QSignalBlocker(self):
            self._slider.wrapped_set_value(value)

        self.sig_update.emit(value)

    def __on_slider_changed(self, value: float) -> None:
        with QtCore.QSignalBlocker(self):
            self.__spin_box.setValue(value)

        self.sig_update.emit(value)


@attrs.frozen(kw_only=True, slots=False)
class OptionalPidgetFactory(PidgetFactory):
    checkbox_label_text: Optional[str] = None


class OptionalPidget(Pidget):
    """Optional parameter, not optional widget"""

    def __init__(self, factory: OptionalPidgetFactory, parent: QWidget) -> None:
        super().__init__(factory, parent)

        self._none_checkbox = QCheckBox(self)
        if factory.checkbox_label_text:
            self._none_checkbox.setText(factory.checkbox_label_text)

    def set_standard_layout(
        self,
        factory: PidgetFactory,
        *,
        first_row_elements: Iterable[
            Union[
                QWidget,
                te.Literal["name_label", "extra_widget", "optional_checkbox"],
            ]
        ],
        full_row_widgets: Collection[QWidget] = (),
        colstretch: tuple[int, ...] = (),
    ) -> None:
        num_optional_checkboxes = sum(e == "optional_checkbox" for e in first_row_elements)
        if num_optional_checkboxes != 1:
            msg = "Exactly 1 'optional_checkbox' should be specified."
            raise ValueError(msg)

        super().set_standard_layout(
            factory,
            first_row_elements=[
                (self._none_checkbox if e == "optional_checkbox" else e)
                for e in first_row_elements
            ],
            full_row_widgets=full_row_widgets,
            colstretch=colstretch,
        )

    def set_data(self, value: Any) -> None:
        with QtCore.QSignalBlocker(self):
            self._none_checkbox.setChecked(value is not None)

    @abc.abstractmethod
    def get_data(self) -> Optional[Any]:
        pass


@attrs.frozen(kw_only=True, slots=False)
class OptionalIntPidgetFactory(OptionalPidgetFactory, IntPidgetFactory):
    init_set_value: Optional[int] = None

    def create(self, parent: QWidget) -> OptionalIntPidget:  # type: ignore[override]
        return OptionalIntPidget(self, parent)


class OptionalIntPidget(OptionalPidget):
    def __init__(self, factory: OptionalIntPidgetFactory, parent: QWidget) -> None:
        super().__init__(factory, parent)

        self._none_checkbox.stateChanged.connect(self.__on_changed)

        self._spin_box = _PidgetSpinBox(
            self,
            limits=factory.limits,
            suffix=factory.suffix,
            init_set_value=factory.init_set_value,
        )
        self._spin_box.valueChanged.connect(self.__on_changed)

        self.set_standard_layout(
            factory,
            first_row_elements=["name_label", "extra_widget", "optional_checkbox", self._spin_box],
            colstretch=(1,),
        )

    def __on_changed(self) -> None:
        checked = self._none_checkbox.isChecked()

        with QtCore.QSignalBlocker(self):
            self._spin_box.setEnabled(checked)

        value = self._spin_box.value() if checked else None
        self.sig_update.emit(value)

    def set_data(self, value: Any) -> None:
        super().set_data(value)

        with QtCore.QSignalBlocker(self):
            if value is None:
                self._spin_box.setEnabled(False)
            else:
                self._spin_box.setValue(value)
                self._spin_box.setEnabled(True)

    def get_data(self) -> Optional[int]:
        if not self._none_checkbox.isChecked():
            return None
        else:
            return int(self._spin_box.value())


@attrs.frozen(kw_only=True, slots=False)
class OptionalFloatPidgetFactory(OptionalPidgetFactory, FloatPidgetFactory):
    init_set_value: Optional[float] = None
    """
    The initial value of the spinbox (ignoring the None-ness of the pidget)
    """

    placeholder_text: Optional[str] = None
    """
    Text that will be displayed when the checkbox is unchecked (and data is None).
    placeholder_text = None disables the placeholder text entierly
    """

    def create(self, parent: QWidget) -> OptionalFloatPidget:  # type: ignore[override]
        return OptionalFloatPidget(self, parent)


class OptionalFloatPidget(OptionalPidget):
    _SPINBOX_INDEX = 0
    _DUMMY_INDEX = 1

    def __init__(self, factory: OptionalFloatPidgetFactory, parent: QWidget) -> None:
        super().__init__(factory, parent)

        self._none_checkbox.stateChanged.connect(self.__on_changed)

        self._container = QStackedWidget()
        self._container.setContentsMargins(0, 0, 0, 0)
        self._spin_box = _PidgetDoubleSpinBox(
            self._container,
            decimals=factory.decimals,
            limits=factory.limits,
            suffix=factory.suffix,
            init_set_value=factory.init_set_value,
        )
        self._spin_box.valueChanged.connect(self.__on_changed)
        self._container.addWidget(self._spin_box)

        if factory.placeholder_text is None:
            self._dummy_spin_box = None
        else:
            self._dummy_spin_box = _PidgetDoubleSpinBox.placeholder_dummy(
                self._container, factory.placeholder_text
            )
            self._dummy_spin_box.setEnabled(False)
            self._container.addWidget(self._dummy_spin_box)

        self.set_standard_layout(
            factory,
            first_row_elements=[
                "name_label",
                "extra_widget",
                "optional_checkbox",
                self._container,
            ],
            colstretch=(1,),
        )

    def __on_changed(self) -> None:
        checked = self._none_checkbox.isChecked()

        with QtCore.QSignalBlocker(self):
            if self._dummy_spin_box is None:
                self._spin_box.setEnabled(checked)
            else:
                self._container.setCurrentIndex(
                    self._SPINBOX_INDEX if checked else self._DUMMY_INDEX
                )

        value = self._spin_box.value() if checked else None
        self.sig_update.emit(value)

    def set_data(self, value: Any) -> None:
        super().set_data(value)

        if value is not None:
            with QtCore.QSignalBlocker(self):
                self._spin_box.setValue(value)

        if self._dummy_spin_box is None:
            self._spin_box.setEnabled(value is not None)
        else:
            self._container.setCurrentIndex(
                self._SPINBOX_INDEX if value is not None else self._DUMMY_INDEX
            )

    def get_data(self) -> Optional[float]:
        if not self._none_checkbox.isChecked():
            return None
        else:
            return float(self._spin_box.value())


@attrs.frozen(kw_only=True, slots=False)
class CheckboxPidgetFactory(PidgetFactory):
    name_label_text: str = attrs.field()

    def create(self, parent: QWidget) -> CheckboxPidget:
        return CheckboxPidget(self, parent)

    @name_label_text.validator
    def check_for_whitespaces(self, attribute: Any, value: str) -> None:
        if value != value.strip():
            msg = "Labels cannot start or end with a whitespace"
            raise ValueError(msg)

    @name_label_text.validator
    def check_label_text_format(self, attribute: Any, value: str) -> None:
        if len(value) > 0 and value[-1] == ":":
            msg = "Checkbox labels cannot end with ':'"
            raise ValueError(msg)


class CheckboxPidget(Pidget):
    def __init__(self, factory: CheckboxPidgetFactory, parent: QWidget) -> None:
        super().__init__(factory, parent)

        self._checkbox = QCheckBox(self)
        self._checkbox.setText(factory.name_label_text)
        self._checkbox.clicked.connect(self.__on_checkbox_click)

        self.set_standard_layout(
            factory,
            first_row_elements=[self._checkbox, "extra_widget"],
            colstretch=(0, 1),
        )

    def __on_checkbox_click(self, checked: bool) -> None:
        self.sig_update.emit(checked)

    def set_data(self, param: Any) -> None:
        with QtCore.QSignalBlocker(self):
            self._checkbox.setChecked(bool(param))

    def get_data(self) -> bool:
        return bool(self._checkbox.isChecked())


@attrs.frozen(kw_only=True, slots=False)
class ComboboxPidgetFactory(PidgetFactory, Generic[T]):
    items: list[tuple[str, T]]

    def create(self, parent: QWidget) -> ComboboxPidget[T]:
        return ComboboxPidget(self, parent)


class ComboboxPidget(Pidget, Generic[T]):
    def __init__(self, factory: ComboboxPidgetFactory[T], parent: QWidget) -> None:
        super().__init__(factory, parent)

        self._combobox = PidgetComboBox(self)

        for displayed_text, user_data in factory.items:
            self._combobox.addItem(displayed_text, user_data)

        self._combobox.currentIndexChanged.connect(self._emit_data_of_combobox_item)

        self.set_standard_layout(
            factory,
            first_row_elements=["name_label", "extra_widget", self._combobox],
        )

    def _emit_data_of_combobox_item(self, index: int) -> None:
        data = self._combobox.itemData(index)
        self.sig_update.emit(data)

    def set_data(self, param: Any) -> None:
        with QtCore.QSignalBlocker(self):
            index = self._combobox.findData(param)
            if index == -1:
                msg = f"Data item {param} could not be found in {self}."
                raise ValueError(msg)
            self._combobox.setCurrentIndex(index)

    def get_data(self) -> T:
        return cast(T, self._combobox.currentData())


@attrs.frozen(kw_only=True, slots=False)
class SensorIdPidgetFactory(ComboboxPidgetFactory[int]):
    name_label_text: str = attrs.field(default="Sensor:")
    name_label_tooltip: str = attrs.field(default="The sensor to use in session")

    def create(self, parent: QWidget) -> SensorIdPidget:
        return SensorIdPidget(self, parent)


class SensorIdPidget(ComboboxPidget[int]):
    def __init__(self, factory: SensorIdPidgetFactory, parent: QWidget) -> None:
        super().__init__(factory, parent)
        self._sensor_id = 1
        self.set_data(self._sensor_id)

    def _emit_data_of_combobox_item(self, index: int) -> None:
        data = self._combobox.itemData(index)

        if data is None:
            return

        self._sensor_id = data
        self.sig_update.emit(data)

    def set_selectable_sensors(self, sensor_list: list[int]) -> None:
        with QtCore.QSignalBlocker(self):
            self._combobox.clear()

            for sensor_id in sensor_list:
                self._combobox.addItem(str(sensor_id), sensor_id)

            self.set_data(self._sensor_id)

    def set_data(self, sensor_id: int) -> None:
        self._sensor_id = sensor_id

        try:
            super().set_data(sensor_id)
        except ValueError:
            with QtCore.QSignalBlocker(self):
                self._combobox.setCurrentIndex(-1)

            self._combobox.setPlaceholderText(f"{self._sensor_id} (unavailable)")

    def get_data(self) -> int:
        return self._sensor_id


@attrs.frozen(kw_only=True, slots=False)
class EnumPidgetFactory(ComboboxPidgetFactory[EnumT]):
    enum_type: Type[EnumT] = attrs.field()
    label_mapping: dict[EnumT, str] = attrs.field()

    items: list[tuple[str, EnumT]] = attrs.field(init=False)

    def __attrs_post_init__(self) -> None:
        if self.label_mapping.keys() != set(self.enum_type):
            msg = "label_mapping does not match enum_type"
            raise ValueError(msg)

        items = [(v, k) for k, v in self.label_mapping.items()]

        # The instance is immutable at this point, which is circumvented by the next row. See:
        # - https://www.attrs.org/en/stable/api.html#attr.ib
        # - https://github.com/python-attrs/attrs/issues/120
        # - https://github.com/python-attrs/attrs/issues/147

        object.__setattr__(self, "items", items)

    def create(self, parent: QWidget) -> EnumPidget[EnumT]:
        return EnumPidget[EnumT](self, parent)


class EnumPidget(ComboboxPidget[EnumT]):
    def __init__(self, factory: EnumPidgetFactory[EnumT], parent: QWidget) -> None:
        super().__init__(factory, parent)


@attrs.frozen(kw_only=True, slots=False)
class OptionalEnumPidgetFactory(OptionalPidgetFactory, EnumPidgetFactory[EnumT]):
    def create(self, parent: QWidget) -> OptionalEnumPidget:  # type: ignore[override]
        return OptionalEnumPidget(self, parent)


class OptionalEnumPidget(OptionalPidget):
    def __init__(self, factory: OptionalEnumPidgetFactory[EnumT], parent: QWidget) -> None:
        super().__init__(factory, parent)

        self._combobox = PidgetComboBox(self)

        for displayed_text, user_data in factory.items:
            self._combobox.addItem(displayed_text, user_data)

        self._none_checkbox.stateChanged.connect(self.__emit_data_of_combobox_or_none)
        self._none_checkbox.stateChanged.connect(self.__enable_combobox_if_checked)
        self._combobox.currentIndexChanged.connect(self.__emit_data_of_combobox_or_none)

        self.set_standard_layout(
            factory,
            first_row_elements=["name_label", "extra_widget", "optional_checkbox", self._combobox],
            colstretch=(1,),
        )

    def __emit_data_of_combobox_or_none(self) -> None:
        if self._none_checkbox.isChecked():
            data = self._combobox.currentData()
            self.sig_update.emit(data)
        else:
            self.sig_update.emit(None)

    def __enable_combobox_if_checked(self) -> None:
        self._combobox.setEnabled(self._none_checkbox.isChecked())

    def set_data(self, value: Any) -> None:
        super().set_data(value)
        if value is not None:
            self.set_enum_parameter(value)
            self._combobox.setEnabled(True)
        else:
            self._combobox.setEnabled(False)

    def get_data(self) -> Optional[Enum]:
        if not self._none_checkbox.isChecked():
            return None
        else:
            return cast(Enum, self._combobox.currentData())

    def set_enum_parameter(self, param: Any) -> None:
        with QtCore.QSignalBlocker(self):
            index = self._combobox.findData(param)
            if index == -1:
                msg = f"Data item {param} could not be found in {self}."
                raise ValueError(msg)
            self._combobox.setCurrentIndex(index)


WIDGET_WIDTH = 125


class PidgetComboBox(QComboBox):
    def __init__(self, parent: QWidget) -> None:
        super().__init__(parent)

        self.setFocusPolicy(QtCore.Qt.FocusPolicy.StrongFocus)
        self.setFixedWidth(WIDGET_WIDTH)

    def wheelEvent(self, event: QtGui.QWheelEvent) -> None:
        if self.hasFocus():
            super().wheelEvent(event)
        else:
            event.ignore()


class _PidgetSpinBox(QSpinBox):
    def __init__(
        self,
        parent: QWidget,
        *,
        limits: Optional[Tuple[Optional[int], Optional[int]]] = None,
        init_set_value: Optional[int] = None,
        suffix: Optional[str] = None,
    ) -> None:
        super().__init__(parent)

        self.setKeyboardTracking(False)
        self.setFocusPolicy(QtCore.Qt.FocusPolicy.StrongFocus)
        self.setAlignment(QtCore.Qt.AlignmentFlag.AlignRight)
        self.setFixedWidth(WIDGET_WIDTH)

        self.setRange(*_convert_int_limits_to_qt_range(limits))

        if suffix:
            self.setSuffix(f" {suffix}")

        if init_set_value is not None:
            self.setValue(init_set_value)

    def wheelEvent(self, event: QtGui.QWheelEvent) -> None:
        if self.hasFocus():
            super().wheelEvent(event)
        else:
            event.ignore()


class _PidgetDoubleSpinBox(QDoubleSpinBox):
    def __init__(
        self,
        parent: QWidget,
        *,
        limits: Optional[Tuple[Optional[float], Optional[float]]] = None,
        init_set_value: Optional[float] = None,
        decimals: int = 1,
        suffix: Optional[str] = None,
    ) -> None:
        super().__init__(parent)

        self.setKeyboardTracking(False)
        self.setFocusPolicy(QtCore.Qt.FocusPolicy.StrongFocus)
        self.setAlignment(QtCore.Qt.AlignmentFlag.AlignRight)
        self.setFixedWidth(WIDGET_WIDTH)

        self.setRange(*_convert_float_limits_to_qt_range(limits))
        self.setDecimals(decimals)
        self.setSingleStep(10 ** (-decimals))

        if suffix:
            self.setSuffix(f" {suffix}")

        if init_set_value is not None:
            self.setValue(init_set_value)

    @classmethod
    def placeholder_dummy(cls, parent: QWidget, placeholder_text: str) -> _PidgetDoubleSpinBox:
        """
        Create a spinbox that only displays a placeholder text
        """
        spin_box = _PidgetDoubleSpinBox(parent)
        spin_box.setSpecialValueText(placeholder_text)
        spin_box.setRange(0, 1)
        spin_box.setValue(0)
        spin_box.setReadOnly(True)
        return spin_box

    def wheelEvent(self, event: QtGui.QWheelEvent) -> None:
        if self.hasFocus():
            super().wheelEvent(event)
        else:
            event.ignore()


class _PidgetFloatSlider(QSlider):
    NUM_STEPS = 1000

    wrapped_value_changed = QtCore.Signal(float)

    def __init__(
        self,
        parent: QWidget,
        *,
        limits: Tuple[float, float],
        decimals: int,
        log_scale: bool,
    ) -> None:
        super().__init__(QtCore.Qt.Orientation.Horizontal, parent)

        self.limits = limits
        self.decimals = decimals
        self.log_scale = log_scale

        self.setRange(0, self.NUM_STEPS)
        self.setFocusPolicy(QtCore.Qt.FocusPolicy.StrongFocus)

        self.valueChanged.connect(self.__on_value_changed)

    def __on_value_changed(self, value: int) -> None:
        self.wrapped_value_changed.emit(self.__from_slider_scale(value))

    def wrapped_set_value(self, value: float) -> None:
        self.setValue(self.__to_slider_scale(value))

    def wheelEvent(self, event: QtGui.QWheelEvent) -> None:
        if self.hasFocus():
            super().wheelEvent(event)
        else:
            event.ignore()

    def __to_slider_scale(self, x: float) -> int:
        lower, upper = self.limits

        if self.log_scale:
            lower = np.log(lower)
            upper = np.log(upper)
            x = np.log(x)

        y = (x - lower) / (upper - lower) * self.NUM_STEPS
        return int(round(y))

    def __from_slider_scale(self, y: int) -> float:
        lower, upper = self.limits

        if self.log_scale:
            lower = np.log(lower)
            upper = np.log(upper)

        x = y / self.NUM_STEPS * (upper - lower) + lower

        if self.log_scale:
            x = np.exp(x)

        return round(x, self.decimals)


def _convert_int_limits_to_qt_range(
    limits: Optional[Tuple[Optional[int], Optional[int]]],
) -> Tuple[int, int]:
    if limits is None:
        limits = (None, None)

    lower, upper = limits

    if lower is None:
        lower = int(-1e9)

    if upper is None:
        upper = int(1e9)

    return (lower, upper)


def _convert_float_limits_to_qt_range(
    limits: Optional[Tuple[Optional[float], Optional[float]]],
) -> Tuple[float, float]:
    if limits is None:
        limits = (None, None)

    lower, upper = limits

    if lower is None:
        lower = -1e9

    if upper is None:
        upper = 1e9

    return (lower, upper)



# Copyright (c) Acconeer AB, 2022-2025
# All rights reserved

from ._bootloader_tool import BootloaderTool
from .xc120_dfu_config import (
    XC120_BOOTLOADER_PID,
    XC120_EXPLORATION_SERVER_WINUSB_PID,
    XC120_VID,
    BootLoaderXC120,
)



# Copyright (c) Acconeer AB, 2022-2024
# All rights reserved

from . import _qt_bindings_canary, a111, utils
from ._core.communication.comm_devices import USBDevice
from ._structs import configbase
from .pg_process import PGProccessDiedException, PGProcess


try:
    # hatch-vcs/setuptools_scm generated version file
    from ._version import __version__
except ImportError as ie:
    msg = "Could not locate version file. Try reinstalling acconeer-exptool."
    raise ImportError(msg) from ie



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved
from __future__ import annotations

import typing as t

from acconeer.exptool import a121
from acconeer.exptool.a121.algo._plugins import processor_main

from . import Processor, ProcessorConfig, get_sensor_config
from ._plugin import PlotPlugin


def _processor_getter(
    session_config: a121.SessionConfig,
    processor_config: ProcessorConfig,
    ignored_metadata: t.Union[a121.Metadata, t.List[t.Dict[int, a121.Metadata]]],
) -> Processor:
    return Processor(
        session_config=session_config,
        processor_config=processor_config,
    )


def _session_config_getter(sensor_id: int) -> a121.SessionConfig:
    return a121.SessionConfig({sensor_id: get_sensor_config()}, extended=True)


processor_main(
    processor_getter=_processor_getter,
    plot_plugin=PlotPlugin,
    processor_config_getter=ProcessorConfig,
    session_config_getter=_session_config_getter,
)



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

import acconeer.exptool as et
from acconeer.exptool import a121

from . import Detector, DetectorConfig


parser = a121.ExampleArgumentParser()
parser.add_argument("--sensor", type=int, default=1)
args = parser.parse_args()
et.utils.config_logging(args)

client = a121.Client.open(**a121.get_client_args(args))

detector_config = DetectorConfig(start_m=0.2, end_m=1)

detector = Detector(
    client=client,
    sensor_ids=[args.sensor],
    detector_config=detector_config,
)

detector.calibrate_detector()
detector.start()

interrupt_handler = et.utils.ExampleInterruptHandler()
print("Press Ctrl-C to end session")

while not interrupt_handler.got_signal:
    detector_result = detector.get_next()

    print(detector_result, "\n")

print("Disconnecting...")
detector.stop()
client.close()



# Copyright (c) Acconeer AB, 2022-2024
# All rights reserved

from __future__ import annotations

import typing as t

import attrs
import numpy as np
import numpy.typing as npt

from acconeer.exptool._core.class_creation.attrs import attrs_ndarray_eq, attrs_ndarray_validator
from acconeer.exptool._core.int_16_complex import INT_16_COMPLEX, int16_complex_array_to_complex

from .metadata import Metadata
from .utils import get_subsweeps_from_frame


@attrs.frozen(kw_only=True)
class ResultContext:
    metadata: Metadata = attrs.field()
    ticks_per_second: int = attrs.field()


@attrs.frozen(kw_only=True)
class Result:
    """Result

    Represents the RSS ``processing_result``.
    """

    data_saturated: bool = attrs.field(converter=bool)
    """
    Indication of sensor data being saturated, can cause data corruption. Lower the receiver gain
    if this indication is set.
    """

    frame_delayed: bool = attrs.field(converter=bool)
    """
    Indication of a delayed frame. The frame rate might need to be lowered if this indication is
    set.
    """

    calibration_needed: bool = attrs.field(converter=bool)
    """
    Indication of calibration needed. The sensor calibration needs to be redone if this indication
    is set.
    """

    temperature: int = attrs.field(converter=int)
    """
    Temperature in sensor during measurement (in degree Celsius). Notice that this has poor
    absolute accuracy.
    """

    _frame: npt.NDArray[t.Any] = attrs.field(
        eq=attrs_ndarray_eq, validator=attrs_ndarray_validator(ndim=2, element_type=INT_16_COMPLEX)
    )
    """Frame data in the original data format (complex int16)"""

    tick: int = attrs.field(converter=int)
    """Server tick when the server got the interrupt from the sensor"""

    _context: ResultContext = attrs.field()

    @property
    def frame(self) -> npt.NDArray[np.complex128]:
        """Frame data in a complex float data format

        2-D with dimensions (sweep, distance).
        """

        return int16_complex_array_to_complex(self._frame)

    @property
    def subframes(self) -> list[npt.NDArray[np.complex128]]:
        """Frame split up into subframes, one for every subsweep config used"""

        return get_subsweeps_from_frame(self.frame, self._context.metadata)

    @property
    def tick_time(self) -> float:
        """Tick converted to a time in seconds"""

        return self.tick / self._context.ticks_per_second



# Copyright (c) Acconeer AB, 2022-2025
# All rights reserved

from __future__ import annotations

import importlib
import logging
import typing as t

import attrs
import typing_extensions as te

from ._enums import PluginFamily
from .app_model import PluginSpec


_REGISTERED_PLUGINS: t.List[PluginSpec] = []
_LOG = logging.getLogger(__name__)


def register_plugin(plugin: PluginSpec) -> None:
    """Registers a plugin, to be used in the Exploration Tool App.

    :param plugin: A plugin
    """
    try:
        plugin = attrs.evolve(plugin, family=PluginFamily.EXTERNAL_PLUGIN)  # type: ignore[misc]
    except Exception:
        _LOG.error(f"Plugin {type(plugin).__name__!r} needs to be an instance of PluginSpecBase")
    finally:
        _REGISTERED_PLUGINS.append(plugin)


def get_registered_plugins() -> list[PluginSpec]:
    return _REGISTERED_PLUGINS


@te.runtime_checkable
class ThirdPartyPluginModule(te.Protocol):
    def register(self) -> None:
        # "self" refers to the module, i.e. is of type "ModuleType"
        ...


def import_and_register_plugin_module(module_name: str) -> None:
    imported_module = importlib.import_module(name=module_name)

    if not isinstance(imported_module, ThirdPartyPluginModule):
        msg = f"Module specified by {module_name!r} does not have a 'register' function."
        raise ValueError(msg)

    try:
        imported_module.register()
    except TypeError:
        raise TypeError(
            f"Function 'register' in {module_name!r} has the wrong signature. "
            + "'register' should have no parameters (should look like this: 'def register(): ...')"
        ) from None


def load_default_plugins() -> list[PluginSpec]:
    from acconeer.exptool.a121.algo.bilateration._plugin import BILATERATION_PLUGIN
    from acconeer.exptool.a121.algo.breathing._ref_app_plugin import BREATHING_PLUGIN
    from acconeer.exptool.a121.algo.cargo._ex_app_plugin import CARGO_PLUGIN
    from acconeer.exptool.a121.algo.distance._detector_plugin import DISTANCE_DETECTOR_PLUGIN
    from acconeer.exptool.a121.algo.hand_motion._example_app_plugin import HAND_MOTION_PLUGIN
    from acconeer.exptool.a121.algo.obstacle._detector_plugin import OBSTACLE_DETECTOR_PLUGIN
    from acconeer.exptool.a121.algo.parking._ref_app_plugin import PARKING_PLUGIN
    from acconeer.exptool.a121.algo.phase_tracking._plugin import PHASE_TRACKING_PLUGIN
    from acconeer.exptool.a121.algo.presence._detector_plugin import PRESENCE_DETECTOR_PLUGIN
    from acconeer.exptool.a121.algo.smart_presence._ref_app_plugin import SMART_PRESENCE_PLUGIN
    from acconeer.exptool.a121.algo.sparse_iq._plugin import SPARSE_IQ_PLUGIN
    from acconeer.exptool.a121.algo.speed._detector_plugin import SPEED_DETECTOR_PLUGIN
    from acconeer.exptool.a121.algo.surface_velocity._example_app_plugin import (
        SURFACE_VELOCITY_PLUGIN,
    )
    from acconeer.exptool.a121.algo.tank_level._plugin import TANK_LEVEL_PLUGIN
    from acconeer.exptool.a121.algo.touchless_button._plugin import TOUCHLESS_BUTTON_PLUGIN
    from acconeer.exptool.a121.algo.vibration._example_app_plugin import VIBRATION_PLUGIN
    from acconeer.exptool.a121.algo.waste_level._plugin import WASTE_LEVEL_PLUGIN

    # Please keep in lexicographical order
    return [
        BILATERATION_PLUGIN,
        BREATHING_PLUGIN,
        CARGO_PLUGIN,
        DISTANCE_DETECTOR_PLUGIN,
        HAND_MOTION_PLUGIN,
        OBSTACLE_DETECTOR_PLUGIN,
        PARKING_PLUGIN,
        PHASE_TRACKING_PLUGIN,
        PRESENCE_DETECTOR_PLUGIN,
        SMART_PRESENCE_PLUGIN,
        SPARSE_IQ_PLUGIN,
        SURFACE_VELOCITY_PLUGIN,
        TANK_LEVEL_PLUGIN,
        TOUCHLESS_BUTTON_PLUGIN,
        VIBRATION_PLUGIN,
        SPEED_DETECTOR_PLUGIN,
        WASTE_LEVEL_PLUGIN,
    ]


def load_plugins() -> list[PluginSpec]:
    return load_default_plugins() + get_registered_plugins()



# Copyright (c) Acconeer AB, 2023-2025
# All rights reserved

from __future__ import annotations

import logging
from enum import Enum, auto
from typing import Callable, List, Mapping, Optional, Tuple

import attrs
import h5py
import numpy as np
import numpy.typing as npt

from PySide6 import QtCore
from PySide6.QtWidgets import QPushButton, QVBoxLayout

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool import a121
from acconeer.exptool._core.docstrings import get_attribute_docstring
from acconeer.exptool.a121._h5_utils import _create_h5_string_dataset
from acconeer.exptool.a121.algo._plugins import (
    A121BackendPluginBase,
    A121ViewPluginBase,
)
from acconeer.exptool.a121.algo.presence._detector import Detector
from acconeer.exptool.a121.algo.presence._detector_plugin import ViewPlugin as PresenceViewPlugin
from acconeer.exptool.app.new import (
    AppModel,
    AttrsConfigEditor,
    BackendLogger,
    GeneralMessage,
    GroupBox,
    Message,
    MiscErrorView,
    PgPlotPlugin,
    PidgetGroupFactoryMapping,
    PluginFamily,
    PluginGeneration,
    PluginPresetBase,
    PluginSpecBase,
    backend,
    icons,
    is_task,
    pidgets,
    visual_policies,
)
from acconeer.exptool.app.new.ui.components import CollapsibleWidget
from acconeer.exptool.app.new.ui.components.a121 import (
    SensorConfigEditor,
)

from ._configs import (
    get_ceiling_config,
    get_long_range_config,
    get_low_power_config,
    get_medium_range_config,
    get_short_range_config,
)
from ._ref_app import (
    PresenceWakeUpConfig,
    PresenceZoneConfig,
    RefApp,
    RefAppConfig,
    RefAppContext,
    RefAppResult,
    _load_algo_data,
    _Mode,
)


log = logging.getLogger(__name__)


@attrs.mutable(kw_only=True)
class SharedState:
    sensor_id: int = attrs.field(default=1)
    config: RefAppConfig = attrs.field(factory=get_medium_range_config)
    ref_app_context: Optional[RefAppContext] = attrs.field(default=None)


class PluginPresetId(Enum):
    SHORT_RANGE = auto()
    MEDIUM_RANGE = auto()
    LONG_RANGE = auto()
    CEILING = auto()
    LOW_POWER = auto()


@attrs.frozen(kw_only=True)
class SetupMessage(GeneralMessage):
    ref_app_config: RefAppConfig
    estimated_frame_rate: float
    nominal_zone_limits: npt.NDArray[np.float64]
    wake_up_zone_limits: npt.NDArray[np.float64]
    name: str = attrs.field(default="setup", init=False)
    recipient: backend.RecipientLiteral = attrs.field(default="plot_plugin", init=False)


class BackendPlugin(A121BackendPluginBase[SharedState]):
    PLUGIN_PRESETS: Mapping[int, Callable[[], RefAppConfig]] = {
        PluginPresetId.SHORT_RANGE.value: lambda: get_short_range_config(),
        PluginPresetId.MEDIUM_RANGE.value: lambda: get_medium_range_config(),
        PluginPresetId.LONG_RANGE.value: lambda: get_long_range_config(),
        PluginPresetId.CEILING.value: lambda: get_ceiling_config(),
        PluginPresetId.LOW_POWER.value: lambda: get_low_power_config(),
    }

    def __init__(
        self, callback: Callable[[Message], None], generation: PluginGeneration, key: str
    ) -> None:
        super().__init__(callback=callback, generation=generation, key=key)

        self._recorder: Optional[a121.H5Recorder] = None
        self._ref_app_instance: Optional[RefApp] = None
        self._log = BackendLogger.getLogger(__name__)

        self.restore_defaults()

    def _load_from_cache(self, file: h5py.File) -> None:
        self.shared_state.config = RefAppConfig.from_json(file["config"][()])

    @is_task
    def restore_defaults(self) -> None:
        self.shared_state = SharedState()
        self.broadcast()

    @is_task
    def update_sensor_id(self, *, sensor_id: int) -> None:
        self.shared_state.sensor_id = sensor_id
        self.broadcast()

    def _sync_sensor_ids(self) -> None:
        if self.client is not None:
            sensor_ids = self.client.server_info.connected_sensors

            if len(sensor_ids) > 0 and self.shared_state.sensor_id not in sensor_ids:
                self.shared_state.sensor_id = sensor_ids[0]

    @is_task
    def update_config(self, *, config: RefAppConfig) -> None:
        self.shared_state.config = config
        self.broadcast()

    @is_task
    def update_nominal_config(self, *, config: PresenceZoneConfig) -> None:
        self.shared_state.config.nominal_config = config
        self.broadcast()

    @is_task
    def update_wake_up_config(self, *, config: PresenceWakeUpConfig) -> None:
        self.shared_state.config.wake_up_config = config
        self.broadcast()

    def save_to_cache(self, file: h5py.File) -> None:
        _create_h5_string_dataset(file, "config", self.shared_state.config.to_json())

    @is_task
    def set_preset(self, preset_id: int) -> None:
        preset_config = self.PLUGIN_PRESETS[preset_id]
        self.shared_state.config = preset_config()
        self.broadcast()

    def load_from_record_setup(self, *, record: a121.H5Record) -> None:
        algo_group = record.get_algo_group(self.key)
        _, config, ref_app_context = _load_algo_data(algo_group)
        self.shared_state.config = config
        self.shared_state.ref_app_context = ref_app_context
        self.shared_state.sensor_id = record.session(0).sensor_id

    def _start_session(self, recorder: Optional[a121.H5Recorder]) -> None:
        assert self.client
        self._ref_app_instance = RefApp(
            client=self.client,
            sensor_id=self.shared_state.sensor_id,
            ref_app_config=self.shared_state.config,
            ref_app_context=self.shared_state.ref_app_context,
        )

        self._ref_app_instance.start(recorder)

        nominal_sensor_config = Detector._get_sensor_config(
            self._ref_app_instance.nominal_detector_config
        )
        distances = np.linspace(
            self.shared_state.config.nominal_config.start_m,
            self.shared_state.config.nominal_config.end_m,
            sum([subsweep.num_points for subsweep in nominal_sensor_config.subsweeps]),
        )
        nominal_zone_limits = self._ref_app_instance.ref_app_processor.create_zones(
            distances, self.shared_state.config.nominal_config.num_zones
        )

        # self._ref_app_instance.ref_app_processor.zone_limits will always be for
        # wake_up_config if wake_up_mode is enabled
        self.callback(
            SetupMessage(
                ref_app_config=self.shared_state.config,
                estimated_frame_rate=self._ref_app_instance.detector.estimated_frame_rate,
                nominal_zone_limits=nominal_zone_limits,
                wake_up_zone_limits=self._ref_app_instance.ref_app_processor.zone_limits,
            )
        )

    def end_session(self) -> None:
        if self._ref_app_instance is None:
            raise RuntimeError
        if self._recorder is not None:
            self._recorder.close()
        self._ref_app_instance.stop()

    def get_next(self) -> None:
        assert self.client
        if self._ref_app_instance is None:
            raise RuntimeError
        result = self._ref_app_instance.get_next()

        self.callback(backend.PlotMessage(result=result))


class PlotPlugin(PgPlotPlugin):
    def __init__(self, app_model: AppModel) -> None:
        super().__init__(app_model=app_model)
        self._plot_job: Optional[RefAppResult] = None
        self._is_setup = False

    def handle_message(self, message: backend.GeneralMessage) -> None:
        if isinstance(message, backend.PlotMessage):
            self._plot_job = message.result
        elif isinstance(message, SetupMessage):
            self.setup(
                message.ref_app_config,
                message.estimated_frame_rate,
                message.nominal_zone_limits,
                message.wake_up_zone_limits,
            )
            self._is_setup = True
        else:
            log.warn(f"{self.__class__.__name__} got an unsupported command: {message.name!r}.")

    def draw(self) -> None:
        if not self._is_setup or self._plot_job is None:
            return

        try:
            self.draw_plot_job(data=self._plot_job)
        finally:
            self._plot_job = None

    def setup(
        self,
        ref_app_config: RefAppConfig,
        estimated_frame_rate: float,
        nominal_zone_limits: npt.NDArray[np.float64],
        wake_up_zone_limits: npt.NDArray[np.float64],
    ) -> None:
        self.plot_layout.clear()

        self.ref_app_config = ref_app_config
        self.nominal_config = ref_app_config.nominal_config
        self.wake_up_config = ref_app_config.wake_up_config

        self.show_all_detected_zones = self.ref_app_config.show_all_detected_zones

        self.history_length_s = 5
        self.time_fifo: List[float] = []
        self.intra_fifo: List[float] = []
        self.inter_fifo: List[float] = []
        self.switch_data: Optional[RefAppResult] = None
        self.switch_data_sectors: List[pg.QtWidgets.QGraphicsEllipseItem] = []

        win = self.plot_layout

        self.intra_limit_lines = []
        self.inter_limit_lines = []

        # Intra presence history plot

        self.intra_hist_plot = win.addPlot(
            row=0,
            col=0,
            title="Intra presence history (fast motions)",
        )
        self.intra_hist_plot.setMenuEnabled(False)
        self.intra_hist_plot.setMouseEnabled(x=False, y=False)
        self.intra_hist_plot.hideButtons()
        self.intra_hist_plot.showGrid(x=True, y=True)
        self.intra_hist_plot.setLabel("bottom", "Time (s)")
        self.intra_hist_plot.setLabel("left", "Score")
        self.intra_hist_plot.setXRange(-self.history_length_s, 0)
        self.intra_history_smooth_max = et.utils.SmoothMax(estimated_frame_rate)
        self.intra_hist_plot.setYRange(0, 10)
        if not self.nominal_config.intra_enable:
            intra_color = et.utils.color_cycler(1)
            intra_color = f"{intra_color}50"
            self.nominal_intra_dashed_pen = pg.mkPen(
                intra_color, width=2.5, style=QtCore.Qt.PenStyle.DashLine
            )
            self.nominal_intra_pen = pg.mkPen(intra_color, width=2)
        else:
            self.nominal_intra_dashed_pen = et.utils.pg_pen_cycler(1, width=2.5, style="--")
            self.nominal_intra_pen = et.utils.pg_pen_cycler(1)

        self.intra_hist_curve = self.intra_hist_plot.plot(pen=self.nominal_intra_pen)
        limit_line = pg.InfiniteLine(angle=0, pen=self.nominal_intra_dashed_pen)
        self.intra_hist_plot.addItem(limit_line)
        self.intra_limit_lines.append(limit_line)

        for line in self.intra_limit_lines:
            line.setPos(self.nominal_config.intra_detection_threshold)

        # Inter presence history plot

        self.inter_hist_plot = win.addPlot(
            row=0,
            col=1,
            title="Inter presence history (slow motions)",
        )
        self.inter_hist_plot.setMenuEnabled(False)
        self.inter_hist_plot.setMouseEnabled(x=False, y=False)
        self.inter_hist_plot.hideButtons()
        self.inter_hist_plot.showGrid(x=True, y=True)
        self.inter_hist_plot.setLabel("bottom", "Time (s)")
        self.inter_hist_plot.setLabel("left", "Score")
        self.inter_hist_plot.setXRange(-self.history_length_s, 0)
        self.inter_history_smooth_max = et.utils.SmoothMax(estimated_frame_rate)
        self.inter_hist_plot.setYRange(0, 10)
        if not self.nominal_config.inter_enable:
            inter_color = et.utils.color_cycler(0)
            inter_color = f"{inter_color}50"
            self.nominal_inter_dashed_pen = pg.mkPen(
                inter_color, width=2.5, style=QtCore.Qt.PenStyle.DashLine
            )
            self.nominal_inter_pen = pg.mkPen(inter_color, width=2)
        else:
            self.nominal_inter_pen = et.utils.pg_pen_cycler(0)
            self.nominal_inter_dashed_pen = et.utils.pg_pen_cycler(0, width=2.5, style="--")

        self.inter_hist_curve = self.inter_hist_plot.plot(pen=self.nominal_inter_pen)
        limit_line = pg.InfiniteLine(angle=0, pen=self.nominal_inter_dashed_pen)
        self.inter_hist_plot.addItem(limit_line)
        self.inter_limit_lines.append(limit_line)

        for line in self.inter_limit_lines:
            line.setPos(self.nominal_config.inter_detection_threshold)

        # Sector plot

        if ref_app_config.wake_up_mode:
            title = (
                "Nominal config<br>"
                "Detection type: fast (orange), slow (blue), both (green)<br>"
                "Green background indicates active"
            )
        else:
            title = "Nominal config<br>" "Detection type: fast (orange), slow (blue), both (green)"

        self.nominal_sector_plot, self.nominal_sectors = self.create_sector_plot(
            title,
            self.ref_app_config.nominal_config.num_zones,
            self.nominal_config.start_m,
            nominal_zone_limits,
        )

        if not ref_app_config.wake_up_mode:
            sublayout = win.addLayout(row=1, col=0, colspan=2)
            sublayout.layout.setColumnStretchFactor(0, 2)
            sublayout.addItem(self.nominal_sector_plot, row=0, col=0)
        else:
            assert self.wake_up_config is not None
            sublayout = win.addLayout(row=1, col=0, colspan=2)
            sublayout.addItem(self.nominal_sector_plot, row=0, col=1)

            title = (
                "Wake up config<br>"
                "Detection type: fast (orange), slow (blue), both (green),<br>"
                "lingering (light grey)<br>"
                "Green background indicates active"
            )
            self.wake_up_sector_plot, self.wake_up_sectors = self.create_sector_plot(
                title,
                self.wake_up_config.num_zones,
                self.wake_up_config.start_m,
                wake_up_zone_limits,
            )

            sublayout.addItem(self.wake_up_sector_plot, row=0, col=0)

            if self.wake_up_config.intra_enable:
                self.wake_up_intra_dashed_pen = et.utils.pg_pen_cycler(1, width=2.5, style="--")
                self.wake_up_intra_pen = et.utils.pg_pen_cycler(1)
            else:
                intra_color = et.utils.color_cycler(1)
                intra_color = f"{intra_color}50"
                self.wake_up_intra_dashed_pen = pg.mkPen(
                    intra_color, width=2.5, style=QtCore.Qt.PenStyle.DashLine
                )
                self.wake_up_intra_pen = pg.mkPen(intra_color, width=2)

            if self.wake_up_config.inter_enable:
                self.wake_up_inter_pen = et.utils.pg_pen_cycler(0)
                self.wake_up_inter_dashed_pen = et.utils.pg_pen_cycler(0, width=2.5, style="--")
            else:
                inter_color = et.utils.color_cycler(0)
                inter_color = f"{inter_color}50"
                self.wake_up_inter_dashed_pen = pg.mkPen(
                    inter_color, width=2.5, style=QtCore.Qt.PenStyle.DashLine
                )
                self.wake_up_inter_pen = pg.mkPen(inter_color, width=2)

    @staticmethod
    def create_sector_plot(
        title: str, num_sectors: int, start_m: float, zone_limits: npt.NDArray[np.float64]
    ) -> Tuple[pg.PlotItem, List[pg.QtWidgets.QGraphicsEllipseItem]]:
        sector_plot = pg.PlotItem(title=title)

        sector_plot.setAspectLocked()
        sector_plot.hideAxis("left")
        sector_plot.hideAxis("bottom")

        sectors = []
        limit_text = []

        range_html = (
            '<div style="text-align: center">'
            '<span style="color: #000000;font-size:12pt;">'
            "{}</span></div>"
        )

        if start_m == zone_limits[0]:
            x_offset = 0.7
        else:
            x_offset = 0

        pen = pg.mkPen("k", width=1)
        span_deg = 25
        for r in np.flip(np.arange(1, num_sectors + 2)):
            sector = pg.QtWidgets.QGraphicsEllipseItem(-r, -r, r * 2, r * 2)
            sector.setStartAngle(-16 * span_deg)
            sector.setSpanAngle(16 * span_deg * 2)
            sector.setPen(pen)
            sector_plot.addItem(sector)
            sectors.append(sector)

            if r != 1:
                limit = pg.TextItem(html=range_html, anchor=(0.5, 0.5), angle=25)
                x = r * np.cos(np.radians(span_deg))
                y = r * np.sin(np.radians(span_deg))
                limit.setPos(x - x_offset, y + 0.25)
                sector_plot.addItem(limit)
                limit_text.append(limit)

        sectors.reverse()

        if start_m != zone_limits[0]:
            start_limit_text = pg.TextItem(html=range_html, anchor=(0.5, 0.5), angle=25)
            start_range_html = range_html.format(f"{start_m}")
            start_limit_text.setHtml(start_range_html)
            x = 1 * np.cos(np.radians(span_deg))
            y = 1 * np.sin(np.radians(span_deg))

            start_limit_text.setPos(x, y + 0.25)
            sector_plot.addItem(start_limit_text)

        unit_text = pg.TextItem(html=range_html, anchor=(0.5, 0.5))
        unit_html = range_html.format("[m]")
        unit_text.setHtml(unit_html)
        x = (num_sectors + 2) * np.cos(np.radians(span_deg))
        y = (num_sectors + 2) * np.sin(np.radians(span_deg))
        unit_text.setPos(x - x_offset, y + 0.25)
        sector_plot.addItem(unit_text)

        for text_item, limit in zip(limit_text, np.flip(zone_limits)):
            zone_range_html = range_html.format(np.around(limit, 1))
            text_item.setHtml(zone_range_html)

        return sector_plot, sectors

    def draw_plot_job(self, *, data: RefAppResult) -> None:
        if data.used_config == _Mode.NOMINAL_CONFIG:
            inter_threshold = self.nominal_config.inter_detection_threshold
            intra_threshold = self.nominal_config.intra_detection_threshold
            intra_pen = self.nominal_intra_pen
            intra_dashed_pen = self.nominal_intra_dashed_pen
            inter_pen = self.nominal_inter_pen
            inter_dashed_pen = self.nominal_inter_dashed_pen
        else:
            assert self.wake_up_config is not None
            inter_threshold = self.wake_up_config.inter_detection_threshold
            intra_threshold = self.wake_up_config.intra_detection_threshold
            intra_pen = self.wake_up_intra_pen
            intra_dashed_pen = self.wake_up_intra_dashed_pen
            inter_pen = self.wake_up_inter_pen
            inter_dashed_pen = self.wake_up_inter_dashed_pen

        self.time_fifo.append(data.service_result.tick_time)

        if data.switch_delay:
            self.intra_fifo.append(float("nan"))
            self.inter_fifo.append(float("nan"))
        else:
            self.intra_fifo.append(data.intra_presence_score)
            self.inter_fifo.append(data.inter_presence_score)

        while self.time_fifo[-1] - self.time_fifo[0] > self.history_length_s:
            self.time_fifo.pop(0)
            self.intra_fifo.pop(0)
            self.inter_fifo.pop(0)

        times = [t - self.time_fifo[-1] for t in self.time_fifo]

        # Intra presence

        if np.isnan(self.intra_fifo).all():
            m_hist = intra_threshold
        else:
            m_hist = np.maximum(float(np.nanmax(self.intra_fifo)), intra_threshold * 1.05)

        m_hist = self.intra_history_smooth_max.update(m_hist)

        self.intra_hist_plot.setYRange(0, m_hist)
        self.intra_hist_curve.setData(times, self.intra_fifo, connect="finite")
        self.intra_hist_curve.setPen(intra_pen)

        for line in self.intra_limit_lines:
            line.setPos(intra_threshold)
            line.setPen(intra_dashed_pen)

        # Inter presence

        if np.isnan(self.inter_fifo).all():
            m_hist = inter_threshold
        else:
            m_hist = np.maximum(float(np.nanmax(self.inter_fifo)), inter_threshold * 1.05)

        m_hist = self.inter_history_smooth_max.update(m_hist)

        self.inter_hist_plot.setYRange(0, m_hist)
        self.inter_hist_curve.setData(times, self.inter_fifo, connect="finite")
        self.inter_hist_curve.setPen(inter_pen)

        for line in self.inter_limit_lines:
            line.setPos(inter_threshold)
            line.setPen(inter_dashed_pen)

        # Sector

        brush = et.utils.pg_brush_cycler(7)
        for sector in self.nominal_sectors:
            sector.setBrush(brush)

        if not self.ref_app_config.wake_up_mode:
            sectors = self.nominal_sectors[1:]
            show_all_zones = self.show_all_detected_zones
            color_nominal = "white"
        else:
            if data.used_config == _Mode.WAKE_UP_CONFIG:
                sectors = self.wake_up_sectors[1:]
                show_all_zones = True
                color_wake_up = "#DFF1D6"
                color_nominal = "white"
            else:
                sectors = self.nominal_sectors[1:]
                show_all_zones = self.show_all_detected_zones
                color_wake_up = "white"
                color_nominal = "#DFF1D6"

            vb = self.nominal_sector_plot.getViewBox()
            vb.setBackgroundColor(color_nominal)
            vb = self.wake_up_sector_plot.getViewBox()
            vb.setBackgroundColor(color_wake_up)

            for sector in self.wake_up_sectors:
                sector.setBrush(brush)

        if data.presence_detected:
            self.color_zones(data, show_all_zones, sectors)
            self.switch_data = data
            self.switch_data_sectors = sectors
        elif data.switch_delay and self.switch_data is not None:
            self.color_zones(self.switch_data, True, self.switch_data_sectors)

        self.nominal_sectors[0].setPen(pg.mkPen(color_nominal, width=1))
        self.nominal_sectors[0].setBrush(pg.mkBrush(color_nominal))

        if self.ref_app_config.wake_up_mode:
            self.wake_up_sectors[0].setPen(pg.mkPen(color_wake_up, width=1))
            self.wake_up_sectors[0].setBrush(pg.mkBrush(color_wake_up))

    @staticmethod
    def color_zones(
        data: RefAppResult,
        show_all_detected_zones: bool,
        sectors: List[pg.QtWidgets.QGraphicsEllipseItem],
    ) -> None:
        if show_all_detected_zones:
            for zone, (inter_value, intra_value) in enumerate(
                zip(data.inter_zone_detections, data.intra_zone_detections)
            ):
                if inter_value + intra_value == 2:
                    sectors[zone].setBrush(et.utils.pg_brush_cycler(2))
                elif inter_value == 1:
                    sectors[zone].setBrush(et.utils.pg_brush_cycler(0))
                elif intra_value == 1:
                    sectors[zone].setBrush(et.utils.pg_brush_cycler(1))
                elif data.used_config == _Mode.WAKE_UP_CONFIG:
                    assert data.wake_up_detections is not None
                    if data.wake_up_detections[zone] > 0:
                        sectors[zone].setBrush(pg.mkBrush("#b5afa0"))
        else:
            assert data.max_presence_zone is not None
            if data.max_presence_zone == data.max_intra_zone:
                sectors[data.max_presence_zone].setBrush(et.utils.pg_brush_cycler(1))
            else:
                sectors[data.max_presence_zone].setBrush(et.utils.pg_brush_cycler(0))


class ViewPlugin(A121ViewPluginBase):
    def __init__(self, app_model: AppModel) -> None:
        super().__init__(app_model=app_model)
        self._log = logging.getLogger(__name__)

        sticky_layout = QVBoxLayout()
        sticky_layout.setContentsMargins(0, 0, 0, 0)
        scrolly_layout = QVBoxLayout()
        scrolly_layout.setContentsMargins(0, 0, 0, 0)

        self.start_button = QPushButton(icons.PLAY(), "Start measurement")
        self.start_button.setShortcut("space")
        self.start_button.setToolTip("Starts the session.\n\nShortcut: Space")
        self.start_button.clicked.connect(self._send_start_request)

        self.stop_button = QPushButton(icons.STOP(), "Stop")
        self.stop_button.setShortcut("space")
        self.stop_button.setToolTip("Stops the session.\n\nShortcut: Space")
        self.stop_button.clicked.connect(self._send_stop_request)

        button_group = GroupBox.grid("Controls", parent=self.sticky_widget)
        button_group.layout().addWidget(self.start_button, 0, 0)
        button_group.layout().addWidget(self.stop_button, 0, 1)

        sticky_layout.addWidget(button_group)

        self.misc_error_view = MiscErrorView(self.scrolly_widget)
        scrolly_layout.addWidget(self.misc_error_view)

        sensor_selection_group = GroupBox.vertical("Sensor selection", parent=self.scrolly_widget)
        self.sensor_id_pidget = pidgets.SensorIdPidgetFactory(items=[]).create(
            parent=sensor_selection_group
        )
        self.sensor_id_pidget.sig_update.connect(self._on_sensor_id_update)
        sensor_selection_group.layout().addWidget(self.sensor_id_pidget)
        scrolly_layout.addWidget(sensor_selection_group)

        self.config_editor = AttrsConfigEditor(
            title="Ref App parameters",
            factory_mapping={
                "wake_up_mode": pidgets.CheckboxPidgetFactory(
                    name_label_text="Switch config on wake up",
                    name_label_tooltip=get_attribute_docstring(RefAppConfig, "wake_up_mode"),
                ),
                "show_all_detected_zones": pidgets.CheckboxPidgetFactory(
                    name_label_text="Show all detected zones",
                    name_label_tooltip=get_attribute_docstring(
                        RefAppConfig, "show_all_detected_zones"
                    ),
                ),
            },
            config_type=RefAppConfig,
            parent=self.scrolly_widget,
        )
        self.config_editor.sig_update.connect(self._on_config_update)
        scrolly_layout.addWidget(self.config_editor)

        self.subsweep_status_wake_up = SensorConfigEditor()
        self.subsweep_status_wake_up.set_read_only(True)
        wake_up_status_widget = CollapsibleWidget(
            "Wake up subsweep settings",
            self.subsweep_status_wake_up.subsweep_group_box,
            self.scrolly_widget,
        )

        scrolly_layout.addWidget(wake_up_status_widget)

        self.wake_up_config_editor = AttrsConfigEditor(
            title="Wake up config parameters",
            factory_mapping=self._get_presence_wake_up_pidget_mapping(),
            config_type=PresenceWakeUpConfig,
            parent=self.scrolly_widget,
        )
        self.wake_up_config_editor.sig_update.connect(self._on_wake_up_config_update)
        scrolly_layout.addWidget(self.wake_up_config_editor)

        self.subsweep_status_nominal = SensorConfigEditor()
        self.subsweep_status_nominal.set_read_only(True)
        nominal_status_widget = CollapsibleWidget(
            "Nominal subsweep settings",
            self.subsweep_status_nominal.subsweep_group_box,
            self.scrolly_widget,
        )

        scrolly_layout.addWidget(nominal_status_widget)
        self.nominal_config_editor = AttrsConfigEditor(
            title="Nominal config parameters",
            factory_mapping=self._get_presence_zone_pidget_mapping(),
            config_type=PresenceZoneConfig,
            parent=self.scrolly_widget,
        )
        self.nominal_config_editor.sig_update.connect(self._on_nominal_config_update)
        scrolly_layout.addWidget(self.nominal_config_editor)

        self.sticky_widget.setLayout(sticky_layout)
        self.scrolly_widget.setLayout(scrolly_layout)

    @classmethod
    def _get_presence_zone_pidget_mapping(cls) -> PidgetGroupFactoryMapping:
        presence_pidget_mapping = dict(PresenceViewPlugin._get_pidget_mapping())
        presence_pidget_mapping.update(
            {
                pidgets.FlatPidgetGroup(): {
                    "num_zones": pidgets.IntPidgetFactory(
                        name_label_text="Number of zones:",
                        limits=(1, None),
                        name_label_tooltip=get_attribute_docstring(
                            PresenceZoneConfig, "num_zones"
                        ),
                    )
                }
            }
        )
        return presence_pidget_mapping

    @classmethod
    def _get_presence_wake_up_pidget_mapping(cls) -> PidgetGroupFactoryMapping:
        presence_pidget_mapping = dict(cls._get_presence_zone_pidget_mapping())
        presence_pidget_mapping.update(
            {
                pidgets.FlatPidgetGroup(): {
                    "num_zones_for_wake_up": pidgets.IntPidgetFactory(
                        name_label_text="Number zones for wake up:s:",
                        limits=(1, None),
                        name_label_tooltip=get_attribute_docstring(
                            PresenceWakeUpConfig, "num_zones_for_wake_up"
                        ),
                    )
                }
            }
        )
        return presence_pidget_mapping

    def on_backend_state_update(self, state: Optional[SharedState]) -> None:
        if state is None:
            self.config_editor.set_data(None)
            self.nominal_config_editor.set_data(None)
            self.subsweep_status_nominal.set_data(None)
            self.wake_up_config_editor.set_data(None)
            self.subsweep_status_wake_up.set_data(None)
        else:
            self.sensor_id_pidget.set_data(state.sensor_id)
            self.config_editor.set_data(state.config)

            self.nominal_config_editor.set_data(state.config.nominal_config)

            self.wake_up_config_editor.setHidden(not state.config.wake_up_mode)

            if state.config.wake_up_config is not None:
                self.wake_up_config_editor.set_data(state.config.wake_up_config)

            results = state.config._collect_validation_results()

            not_handled = self.config_editor.handle_validation_results(results)

            not_handled = self.wake_up_config_editor.handle_validation_results(not_handled)

            not_handled = self.misc_error_view.handle_validation_results(not_handled)

            assert not_handled == []

            if len(results) == 0:
                self.subsweep_status_nominal.set_data(
                    Detector._get_sensor_config(state.config.nominal_config)
                )
                if state.config.wake_up_config is not None:
                    self.subsweep_status_wake_up.set_data(
                        Detector._get_sensor_config(state.config.wake_up_config)
                    )

    def on_app_model_update(self, app_model: AppModel) -> None:
        self.sensor_id_pidget.set_selectable_sensors(app_model.connected_sensors)

        visual_policies.apply_enabled_policy(
            visual_policies.config_editor_enabled,
            app_model,
            widgets=[
                self.config_editor,
                self.nominal_config_editor,
                self.wake_up_config_editor,
                self.sensor_id_pidget,
            ],
        )

        self.start_button.setEnabled(
            visual_policies.start_button_enabled(
                app_model, extra_condition=self.config_editor.is_ready
            )
        )
        self.stop_button.setEnabled(visual_policies.stop_button_enabled(app_model))

    def _on_config_update(self, config: RefAppConfig) -> None:
        BackendPlugin.update_config.rpc(self.app_model.put_task, config=config)

    def _on_nominal_config_update(self, config: PresenceZoneConfig) -> None:
        BackendPlugin.update_nominal_config.rpc(self.app_model.put_task, config=config)

    def _on_wake_up_config_update(self, config: PresenceWakeUpConfig) -> None:
        BackendPlugin.update_wake_up_config.rpc(self.app_model.put_task, config=config)

    def _send_defaults_request(self) -> None:
        BackendPlugin.restore_defaults.rpc(self.app_model.put_task)

    def _on_sensor_id_update(self, sensor_id: int) -> None:
        BackendPlugin.update_sensor_id.rpc(self.app_model.put_task, sensor_id=sensor_id)


class PluginSpec(PluginSpecBase):
    def create_backend_plugin(
        self, callback: Callable[[Message], None], key: str
    ) -> BackendPlugin:
        return BackendPlugin(callback=callback, generation=self.generation, key=key)

    def create_view_plugin(self, app_model: AppModel) -> ViewPlugin:
        return ViewPlugin(app_model=app_model)

    def create_plot_plugin(self, app_model: AppModel) -> PlotPlugin:
        return PlotPlugin(app_model=app_model)


SMART_PRESENCE_PLUGIN = PluginSpec(
    generation=PluginGeneration.A121,
    key="smart_presence",
    title="Smart presence",
    docs_link="https://docs.acconeer.com/en/latest/ref_apps/a121/smart_presence.html",
    description="Split presence detection range into zones.",
    family=PluginFamily.REF_APP,
    presets=[
        PluginPresetBase(
            name="Short range",
            description="Short range",
            preset_id=PluginPresetId.SHORT_RANGE,
        ),
        PluginPresetBase(
            name="Medium range",
            description="Medium range",
            preset_id=PluginPresetId.MEDIUM_RANGE,
        ),
        PluginPresetBase(
            name="Long range",
            description="Long range",
            preset_id=PluginPresetId.LONG_RANGE,
        ),
        PluginPresetBase(
            name="Ceiling",
            description="Ceiling",
            preset_id=PluginPresetId.CEILING,
        ),
        PluginPresetBase(
            name="Low power, Wake up",
            description="Low power, Wake up",
            preset_id=PluginPresetId.LOW_POWER,
        ),
    ],
    default_preset_id=PluginPresetId.MEDIUM_RANGE,
)
